
import tensorflow as tf
import numpy as np
import random

class Graph():
	def __init__(self, nx_G, is_directed, p, q):
		self.G = nx_G
		self.is_directed = is_directed
		self.p = p
		self.q = q

	def node2vec_walk(self, walk_length, start_node):
		'''
		Simulate a random walk starting from start node.
		'''
		G = self.G
		alias_nodes = self.alias_nodes
		alias_edges = self.alias_edges

		walk = [start_node]

		while len(walk) < walk_length:
			cur = walk[-1]
			cur_nbrs = sorted(G.neighbors(cur))
			if len(cur_nbrs) > 0:
				if len(walk) == 1:
					walk.append(cur_nbrs[alias_draw(alias_nodes[cur][0], alias_nodes[cur][1])])
				else:
					prev = walk[-2]
					next = cur_nbrs[alias_draw(alias_edges[(prev, cur)][0],
						alias_edges[(prev, cur)][1])]
					walk.append(next)
			else:
				break

		return walk

	def simulate_walks(self, num_walks, walk_length):
		'''
		Repeatedly simulate random walks from each node.
		'''
		G = self.G
		walks = []
		nodes = list(G.nodes())
		print ('Walk iteration:')
		for walk_iter in range(num_walks):
			print (str(walk_iter+1), '/', str(num_walks))
			random.shuffle(nodes)
			for node in nodes:
				walks.append(self.node2vec_walk(walk_length=walk_length, start_node=node))
		return walks

	def get_alias_edge(self, src, dst):
		'''
		Get the alias edge setup lists for a given edge.
		'''
		G = self.G
		p = self.p
		q = self.q
		unnormalized_probs = []
		for dst_nbr in sorted(G.neighbors(dst)):
			if dst_nbr == src:
				unnormalized_probs.append(G[dst][dst_nbr]['weight']/p)
			elif G.has_edge(dst_nbr, src):
				unnormalized_probs.append(G[dst][dst_nbr]['weight'])
			else:
				unnormalized_probs.append(G[dst][dst_nbr]['weight']/q)
		norm_const = sum(unnormalized_probs)
		normalized_probs =  [float(u_prob)/norm_const for u_prob in unnormalized_probs]

		return alias_setup(normalized_probs)

	def preprocess_transition_probs(self):
		'''
		Preprocessing of transition probabilities for guiding the random walks.
		'''
		G = self.G
		is_directed = self.is_directed
		alias_nodes = {}
		for node in G.nodes():
			unnormalized_probs = [G[node][nbr]['weight'] for nbr in sorted(G.neighbors(node))]
			norm_const = sum(unnormalized_probs)
			if norm_const == 0:
				print("node:",node)
				print("unnormalized_probs:",unnormalized_probs)
			normalized_probs =  [float(u_prob)/norm_const for u_prob in unnormalized_probs]
			alias_nodes[node] = alias_setup(normalized_probs)

		alias_edges = {}
		triads = {}

		if is_directed:
			for edge in G.edges():
				alias_edges[edge] = self.get_alias_edge(edge[0], edge[1])
		else:
			for edge in G.edges():
				alias_edges[edge] = self.get_alias_edge(edge[0], edge[1])
				alias_edges[(edge[1], edge[0])] = self.get_alias_edge(edge[1], edge[0])

		self.alias_nodes = alias_nodes
		self.alias_edges = alias_edges

		return

def alias_setup(probs):
	'''
	Compute utility lists for non-uniform sampling from discrete distributions.
	Refer to https://hips.seas.harvard.edu/blog/2013/03/03/the-alias-method-efficient-sampling-with-many-discrete-outcomes/
	for details
	'''
	K = len(probs)
	q = np.zeros(K)
	J = np.zeros(K, dtype=np.int)

	smaller = []
	larger = []
	for kk, prob in enumerate(probs):
		q[kk] = K*prob
		if q[kk] < 1.0:
			smaller.append(kk)
		else:
			larger.append(kk)

	while len(smaller) > 0 and len(larger) > 0:
		small = smaller.pop()
		large = larger.pop()

		J[small] = large
		q[large] = q[large] + q[small] - 1.0
		if q[large] < 1.0:
			smaller.append(large)
		else:
			larger.append(large)

	return J, q

def alias_draw(J, q):
	'''
	Draw sample from a non-uniform discrete distribution using alias sampling.
	'''
	K = len(J)

	kk = int(np.floor(np.random.rand()*K))
	if np.random.rand() < q[kk]:
		return kk
	else:
		return J[kk]


def conv2d(x, output_dims, kernel_size, stride = [1, 1],
		   padding = 'SAME', use_bias = True, activation = tf.nn.relu,
		   bn = False, bn_decay = None, is_training = None):
	input_dims = x.get_shape()[-1].value
	kernel_shape = kernel_size + [input_dims, output_dims]
	kernel = tf.Variable(
		tf.glorot_uniform_initializer()(shape = kernel_shape),
		dtype = tf.float32, trainable = True, name = 'kernel')
	x = tf.nn.conv2d(x, kernel, [1] + stride + [1], padding = padding)
	if use_bias:
		bias = tf.Variable(
			tf.zeros_initializer()(shape = [output_dims]),
			dtype = tf.float32, trainable = True, name = 'bias')
		x = tf.nn.bias_add(x, bias)
	if activation is not None:
		if bn:
			x = batch_norm(x, is_training = is_training, bn_decay = bn_decay)
		x = activation(x)
	return x

def batch_norm(x, is_training, bn_decay):
	input_dims = x.get_shape()[-1].value
	moment_dims = list(range(len(x.get_shape()) - 1))
	beta = tf.Variable(
		tf.zeros_initializer()(shape = [input_dims]),
		dtype = tf.float32, trainable = True, name = 'beta')
	gamma = tf.Variable(
		tf.ones_initializer()(shape = [input_dims]),
		dtype = tf.float32, trainable = True, name = 'gamma')
	batch_mean, batch_var = tf.nn.moments(x, moment_dims, name='moments')

	decay = bn_decay if bn_decay is not None else 0.9
	ema = tf.train.ExponentialMovingAverage(decay = decay)
	# Operator that maintains moving averages of variables.
	ema_apply_op = tf.cond(
		is_training,
		lambda: ema.apply([batch_mean, batch_var]),
		lambda: tf.no_op())
	# Update moving average and return current batch's avg and var.
	def mean_var_with_update():
		with tf.control_dependencies([ema_apply_op]):
			return tf.identity(batch_mean), tf.identity(batch_var)
	# ema.average returns the Variable holding the average of var.
	mean, var = tf.cond(
		is_training,
		mean_var_with_update,
		lambda: (ema.average(batch_mean), ema.average(batch_var)))
	x = tf.nn.batch_normalization(x, mean, var, beta, gamma, 1e-3)
	return x

def dropout(x, drop, is_training):
	x = tf.cond(
		is_training,
		lambda: tf.nn.dropout(x, rate = drop),
		lambda: x)
	return x


def placeholder(P, Q, N):
	X = tf.compat.v1.placeholder(shape=(None, P, N), dtype=tf.float32)
	TE = tf.compat.v1.placeholder(shape=(None, P + Q, 2), dtype=tf.int32)
	label = tf.compat.v1.placeholder(shape=(None, Q, N), dtype=tf.float32)
	is_training = tf.compat.v1.placeholder(shape=(), dtype=tf.bool)
	return X, TE, label, is_training


def FC(x, units, activations, bn, bn_decay, is_training, use_bias=True):
	if isinstance(units, int):
		units = [units]
		activations = [activations]
	elif isinstance(units, tuple):
		units = list(units)
		activations = list(activations)
	assert type(units) == list
	for num_unit, activation in zip(units, activations):
		x = conv2d(
			x, output_dims=num_unit, kernel_size=[1, 1], stride=[1, 1],
			padding='VALID', use_bias=use_bias, activation=activation,
			bn=bn, bn_decay=bn_decay, is_training=is_training)
	return x


def STEmbedding(SE, TE, T, D, bn, bn_decay, is_training):
	'''
	spatio-temporal embedding
	SE:	 [N, D]
	TE:	 [batch_size, P + Q, 2] (dayofweek, timeofday)
	T:	  num of time steps in one day
	D:	  output dims
	retrun: [batch_size, P + Q, N, D]
	'''
	# spatial embedding
	SE = tf.expand_dims(tf.expand_dims(SE, axis=0), axis=0)
	SE = FC(
		SE, units=[D, D], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	# temporal embedding
	dayofweek = tf.one_hot(TE[..., 0], depth=7)
	timeofday = tf.one_hot(TE[..., 1], depth=T)
	TE = tf.concat((dayofweek, timeofday), axis=-1)
	TE = tf.expand_dims(TE, axis=2)
	TE = FC(
		TE, units=[D, D], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	return tf.add(SE, TE)


def spatialAttention(X, STE, K, d, bn, bn_decay, is_training):
	'''
	spatial attention mechanism
	X:	  [batch_size, num_step, N, D]
	STE:	[batch_size, num_step, N, D]
	K:	  number of attention heads
	d:	  dimension of each attention outputs
	return: [batch_size, num_step, N, D]
	'''
	D = K * d
	X = tf.concat((X, STE), axis=-1)
	# [batch_size, num_step, N, K * d]
	query = FC(
		X, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	key = FC(
		X, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	value = FC(
		X, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	# [K * batch_size, num_step, N, d]
	query = tf.concat(tf.split(query, K, axis=-1), axis=0)
	key = tf.concat(tf.split(key, K, axis=-1), axis=0)
	value = tf.concat(tf.split(value, K, axis=-1), axis=0)
	# [K * batch_size, num_step, N, N]
	attention = tf.matmul(query, key, transpose_b=True)
	attention /= (d ** 0.5)
	attention = tf.nn.softmax(attention, axis=-1)
	# [batch_size, num_step, N, D]
	X = tf.matmul(attention, value)
	X = tf.concat(tf.split(X, K, axis=0), axis=-1)
	X = FC(
		X, units=[D, D], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	return X


def temporalAttention(X, STE, K, d, bn, bn_decay, is_training, mask=True):
	'''
	temporal attention mechanism
	X:	  [batch_size, num_step, N, D]
	STE:	[batch_size, num_step, N, D]
	K:	  number of attention heads
	d:	  dimension of each attention outputs
	return: [batch_size, num_step, N, D]
	'''
	D = K * d
	X = tf.concat((X, STE), axis=-1)
	# [batch_size, num_step, N, K * d]
	query = FC(
		X, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	key = FC(
		X, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	value = FC(
		X, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	# [K * batch_size, num_step, N, d]
	query = tf.concat(tf.split(query, K, axis=-1), axis=0)
	key = tf.concat(tf.split(key, K, axis=-1), axis=0)
	value = tf.concat(tf.split(value, K, axis=-1), axis=0)
	# query: [K * batch_size, N, num_step, d]
	# key:   [K * batch_size, N, d, num_step]
	# value: [K * batch_size, N, num_step, d]
	query = tf.transpose(query, perm=(0, 2, 1, 3))
	key = tf.transpose(key, perm=(0, 2, 3, 1))
	value = tf.transpose(value, perm=(0, 2, 1, 3))
	# [K * batch_size, N, num_step, num_step]
	attention = tf.matmul(query, key)
	attention /= (d ** 0.5)
	# mask attention score
	if mask:
		batch_size = tf.shape(X)[0]
		num_step = X.get_shape()[1].value
		N = X.get_shape()[2].value
		mask = tf.ones(shape=(num_step, num_step))
		mask = tf.linalg.LinearOperatorLowerTriangular(mask).to_dense()
		mask = tf.expand_dims(tf.expand_dims(mask, axis=0), axis=0)
		mask = tf.tile(mask, multiples=(K * batch_size, N, 1, 1))
		mask = tf.cast(mask, dtype=tf.bool)
		attention = tf.compat.v2.where(
			condition=mask, x=attention, y=-2 ** 15 + 1)
	# softmax
	attention = tf.nn.softmax(attention, axis=-1)
	# [batch_size, num_step, N, D]
	X = tf.matmul(attention, value)
	X = tf.transpose(X, perm=(0, 2, 1, 3))
	X = tf.concat(tf.split(X, K, axis=0), axis=-1)
	X = FC(
		X, units=[D, D], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	return X


def gatedFusion(HS, HT, D, bn, bn_decay, is_training):
	'''
	gated fusion
	HS:	 [batch_size, num_step, N, D]
	HT:	 [batch_size, num_step, N, D]
	D:	  output dims
	return: [batch_size, num_step, N, D]
	'''
	XS = FC(
		HS, units=D, activations=None,
		bn=bn, bn_decay=bn_decay,
		is_training=is_training, use_bias=False)
	XT = FC(
		HT, units=D, activations=None,
		bn=bn, bn_decay=bn_decay,
		is_training=is_training, use_bias=True)
	z = tf.nn.sigmoid(tf.add(XS, XT))
	H = tf.add(tf.multiply(z, HS), tf.multiply(1 - z, HT))
	H = FC(
		H, units=[D, D], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	return H


def STAttBlock(X, STE, K, d, bn, bn_decay, is_training, mask=False):
	HS = spatialAttention(X, STE, K, d, bn, bn_decay, is_training)
	HT = temporalAttention(X, STE, K, d, bn, bn_decay, is_training, mask=mask)
	H = gatedFusion(HS, HT, K * d, bn, bn_decay, is_training)
	return tf.add(X, H)


def transformAttention(X, STE_P, STE_Q, K, d, bn, bn_decay, is_training):
	'''
	transform attention mechanism
	X:	  [batch_size, P, N, D]
	STE_P:  [batch_size, P, N, D]
	STE_Q:  [batch_size, Q, N, D]
	K:	  number of attention heads
	d:	  dimension of each attention outputs
	return: [batch_size, Q, N, D]
	'''
	D = K * d
	# query: [batch_size, Q, N, K * d]
	# key:   [batch_size, P, N, K * d]
	# value: [batch_size, P, N, K * d]
	query = FC(
		STE_Q, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	key = FC(
		STE_P, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	value = FC(
		X, units=D, activations=tf.nn.relu,
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	# query: [K * batch_size, Q, N, d]
	# key:   [K * batch_size, P, N, d]
	# value: [K * batch_size, P, N, d]
	query = tf.concat(tf.split(query, K, axis=-1), axis=0)
	key = tf.concat(tf.split(key, K, axis=-1), axis=0)
	value = tf.concat(tf.split(value, K, axis=-1), axis=0)
	# query: [K * batch_size, N, Q, d]
	# key:   [K * batch_size, N, d, P]
	# value: [K * batch_size, N, P, d]
	query = tf.transpose(query, perm=(0, 2, 1, 3))
	key = tf.transpose(key, perm=(0, 2, 3, 1))
	value = tf.transpose(value, perm=(0, 2, 1, 3))
	# [K * batch_size, N, Q, P]
	attention = tf.matmul(query, key)
	attention /= (d ** 0.5)
	attention = tf.nn.softmax(attention, axis=-1)
	# [batch_size, Q, N, D]
	X = tf.matmul(attention, value)
	X = tf.transpose(X, perm=(0, 2, 1, 3))
	X = tf.concat(tf.split(X, K, axis=0), axis=-1)
	X = FC(
		X, units=[D, D], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	return X


def GMAN(X, TE, SE, P, Q, T, L, K, d, bn, bn_decay, is_training):
	"""
	References:
        - `Gman: A graph multi-attention network for traffic prediction.
          <https://ojs.aaai.org/index.php/AAAI/article/view/5477>`_.
        - `A Tensorflow implementation of the GMAN model  (Zhengchuanpan)
          <https://github.com/zhengchuanpan/GMAN>`_.

    Args:
		P(int): Number of history steps.
		Q(int): Number of prediction steps.
		T(int): Number of steps which one day is divided into.
		L(int): Number of STAtt blocks in the encoder/decoder.
		K(int): Number of attention heads.
		d(int): Number of dimension of each attention head outputs.
        X(tf.Tensor): Input traffic data with shape [batch_size, ...]
        TE(tf.Tensor): Temporal embedding [batch_size, ...]
        SE(tf.Tensor): Spatial embedding [batch_size, ...]
		bn(bool): Whether to do batch normalization.
		is_training(bool): Whether to train.
    """
	D = K * d
	# input
	X = tf.expand_dims(X, axis=-1)
	X = FC(
		X, units=[D, D], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	# STE
	STE = STEmbedding(SE, TE, T, D, bn, bn_decay, is_training)
	STE_P = STE[:, : P]
	STE_Q = STE[:, P:]
	# encoder
	for _ in range(L):
		X = STAttBlock(X, STE_P, K, d, bn, bn_decay, is_training)
	# transAtt
	X = transformAttention(
		X, STE_P, STE_Q, K, d, bn, bn_decay, is_training)
	# decoder
	for _ in range(L):
		X = STAttBlock(X, STE_Q, K, d, bn, bn_decay, is_training)
	# output
	X = FC(
		X, units=[D, 1], activations=[tf.nn.relu, None],
		bn=bn, bn_decay=bn_decay, is_training=is_training)
	return tf.squeeze(X, axis=3)




import os
import tensorflow as tf

from ..model_unit import BaseModel


class DeepST(BaseModel):
    """Deep learning-based prediction model for Spatial-Temporal data (DeepST)

    DeepST is composed of three components: 1) temporal dependent
    instances: describing temporal closeness, period and seasonal
    trend; 2) convolutional neural networks: capturing near and
    far spatial dependencies; 3) early and late fusions: fusing
    similar and different domains' data.

    Reference:
        - `DNN-Based Prediction Model for Spatial-Temporal Data (Junbo Zhang et al., 2016)
          <https://www.microsoft.com/en-us/research/wp-content/uploads/2016/09/DeepST-SIGSPATIAL2016.pdf>`_.

    Args:
        closeness_len (int): The length of closeness data history. The former consecutive ``closeness_len`` time slots
            of data will be used as closeness history.
        period_len (int): The length of period data history. The data of exact same time slots in former consecutive
            ``period_len`` days will be used as period history.
        trend_len (int): The length of trend data history. The data of exact same time slots in former consecutive
            ``trend_len`` weeks (every seven days) will be used as trend history.
        width (int): The width of grid data.
        height (int): The height of grid data.
        externai_dim (int): Number of dimensions of external data.
        kernel_size (int): Kernel size in Convolutional neural networks. Default: 3
        num_conv_filters (int):  the Number of filters in the convolution. Default: 64
        lr (float): Learning rate. Default: 1e-5
        code_version (str): Current version of this model code.
        model_dir (str): The directory to store model files. Default:'model_dir'
        gpu_device (str): To specify the GPU to use. Default: '0'
    """
    def __init__(self,
                 closeness_len,
                 period_len,
                 trend_len,
                 width,
                 height,
                 external_dim,
                 kernel_size=3,
                 num_conv_filters=64,
                 lr=1e-5,
                 code_version='QuickStart-DeepST',
                 model_dir='model_dir',
                 gpu_device='0'):
        
        super(DeepST, self).__init__(code_version=code_version, model_dir=model_dir, gpu_device=gpu_device)

        self._width = width
        self._height = height

        self._closeness_len = closeness_len
        self._period_len = period_len
        self._trend_len = trend_len

        self._external_dim = external_dim
        self._lr = lr
        self._kernel_size = kernel_size
        self._num_conv_filters = num_conv_filters

        self._graph = tf.Graph()
        self._GPU_DEVICE = gpu_device

        self._input = {}
        self._output = {}
        self._op = {}
        self._variable_init = None
        self._saver = None
        self._model_dir = model_dir

        # GPU Config
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = self._GPU_DEVICE
        self._config = tf.ConfigProto()
        self._config.gpu_options.allow_growth = True

        self._session = tf.Session(graph=self._graph, config=self._config)
    
    def build(self):
        with self._graph.as_default():

            c_conf = tf.placeholder(tf.float32, [None, self._height, self._width, self._closeness_len], name='c')
            p_conf = tf.placeholder(tf.float32, [None, self._height, self._width, self._period_len], name='p')
            t_conf = tf.placeholder(tf.float32, [None, self._height, self._width, self._trend_len], name='t')

            target = tf.placeholder(tf.float32, [None, self._height, self._width, 1], name='target')
            
            self._input['closeness_feature'] = c_conf.name
            self._input['period_feature'] = p_conf.name
            self._input['trend_feature'] = t_conf.name
            self._input['target'] = target.name

            # First convolution
            h_c_1 = tf.layers.conv2d(inputs=c_conf, filters=self._num_conv_filters,
                                     kernel_size=[self._kernel_size, self._kernel_size], padding='SAME', use_bias=True)
            h_p_1 = tf.layers.conv2d(inputs=c_conf, filters=self._num_conv_filters,
                                     kernel_size=[self._kernel_size, self._kernel_size], padding='SAME', use_bias=True)
            h_t_1 = tf.layers.conv2d(inputs=c_conf, filters=self._num_conv_filters,
                                     kernel_size=[self._kernel_size, self._kernel_size], padding='SAME', use_bias=True)

            # First fusion
            h_2 = tf.layers.conv2d(tf.concat([h_c_1, h_p_1, h_t_1], axis=-1),
                                   filters=self._num_conv_filters, kernel_size=[self._kernel_size, self._kernel_size],
                                   padding='SAME', use_bias=True)

            # Stack more convolutions
            middle_output = tf.layers.conv2d(h_2, filters=self._num_conv_filters,
                                             kernel_size=[self._kernel_size, self._kernel_size],
                                             padding='SAME', use_bias=True)
            x = tf.layers.conv2d(middle_output, filters=self._num_conv_filters,
                                 kernel_size=[self._kernel_size, self._kernel_size], padding='SAME', use_bias=True)

            # external dims
            if self._external_dim is not None and self._external_dim > 0:
                external_input = tf.placeholder(tf.float32, [None, self._external_dim])
                self._input['external_feature'] = external_input.name
                external_dense = tf.layers.dense(inputs=external_input, units=10)
                external_dense = tf.tile(tf.reshape(external_dense, [-1, 1, 1, 10]), [1, self._height, self._width, 1])
                x = tf.concat([x, external_dense], axis=-1)

            x = tf.layers.dense(x, units=1, name='prediction', activation=tf.nn.sigmoid)

            loss = tf.sqrt(tf.reduce_mean(tf.square(x - target)), name='loss')
            train_op = tf.train.AdamOptimizer(self._lr).minimize(loss)

            self._output['prediction'] = x.name
            self._output['loss'] = loss.name

            self._op['train_op'] = train_op.name

            self._variable_init = tf.global_variables_initializer()
            self._saver = tf.train.Saver()

        super(DeepST, self).build()

    # Define your '_get_feed_dict function‘, map your input to the tf-model
    def _get_feed_dict(self, closeness_feature=None, period_feature=None, trend_feature=None,
                       target=None, external_feature=None):
        '''
        The method to get feet dict for tensorflow model.

        Users may modify this method according to the format of input.

        Args:
            closeness_feature (np.ndarray or ``None``): The closeness history data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, closeness_len].
            period_feature (np.ndarray or ``None``): The period history data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, period_len].
            trend_feature (np.ndarray or ``None``): The trend history data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, trend_len].
            target (np.ndarray or ``None``): The target value data.
                If type is np.ndarray, its shape is [time_slot_num, height, width, 1].
            external_feature (np.ndarray or ``None``): The external feature data.
                If type is np.ndaaray, its shape is [time_slot_num, feature_num].
        '''
        feed_dict = {}
        if target is not None:
            feed_dict['target'] = target
        if self._external_dim is not None and self._external_dim > 0:
            feed_dict['external_feature'] = external_feature
        if self._closeness_len is not None and self._closeness_len > 0:
            feed_dict['closeness_feature'] = closeness_feature
        if self._period_len is not None and self._period_len > 0:
            feed_dict['period_feature'] = period_feature
        if self._trend_len is not None and self._trend_len > 0:
            feed_dict['trend_feature'] = trend_feature
        return feed_dict



import tensorflow as tf

from ..model_unit import BaseModel
from ..model_unit.GraphModelLayers import GCL


class ST_MGCN(BaseModel):
    """

    References:
        - `Spatiotemporal multi-graph convolution network for ride-hailing demand forecasting (Geng Xu, et al., 2019)
          <http://www-scf.usc.edu/~yaguang/papers/aaai19_multi_graph_convolution.pdf>`_.
        - `A PyTorch implementation of the ST-MGCN model  (shawnwang-tech)
          <https://github.com/shawnwang-tech/ST-MGCN-pytorch>`_.

    Args:
        T(int): Input sequence length
        input_dim(int): Input feature dimension
        num_graph(int): Number of graphs used in the model.
        gcl_k(int): The highest order of Chebyshev Polynomial approximation in GCN.
        gcl_l(int): Number of GCN layers.
        lstm_units(int): Number of hidden units of RNN.
        lstm_layers(int): Number of LSTM layers.
        lr(float): Learning rate
        external_dim(int): Dimension of the external feature, e.g. temperature and wind are two dimension.
        code_version(str): Current version of this model code, which will be used as filename for saving the model
        model_dir(str): The directory to store model files. Default:'model_dir'.
        gpu_device(str): To specify the GPU to use. Default: '0'.
    """
    def __init__(self,
                 T,
                 input_dim,
                 num_graph,
                 gcl_k,
                 gcl_l,
                 lstm_units,
                 lstm_layers,
                 lr,
                 external_dim,
                 code_version,
                 model_dir,
                 gpu_device):

        super(ST_MGCN, self).__init__(code_version=code_version,
                                      model_dir=model_dir,
                                      gpu_device=gpu_device)

        self._T = T
        self._input_dim = input_dim
        self._num_graph = num_graph
        self._gcl_k = gcl_k
        self._gcl_l = gcl_l
        self._lstm_units = lstm_units
        self._lstm_layers = lstm_layers
        self._lr = lr
        self._external_dim = external_dim

    def build(self, init_vars=True, max_to_keep=5):
        with self._graph.as_default():
            # [batch, T, num_stations, input_dim]
            traffic_flow = tf.placeholder(tf.float32, [None, self._T, None, self._input_dim])
            laplacian_matrix = tf.placeholder(tf.float32, [self._num_graph, None, None])
            target = tf.placeholder(tf.float32, [None, None, 1])

            self._input['traffic_flow'] = traffic_flow.name
            self._input['laplace_matrix'] = laplacian_matrix.name
            self._input['target'] = target.name

            station_number = tf.shape(traffic_flow)[-2]

            outputs = []

            for graph_index in range(self._num_graph):
                with tf.variable_scope('CGRNN_Graph%s' % graph_index, reuse=False):
                    f_k_g = GCL.add_multi_gc_layers(tf.reshape(traffic_flow, [-1, station_number, self._input_dim]),
                                                    gcn_k=1, gcn_l=1, output_size=self._input_dim,
                                                    laplacian_matrix=laplacian_matrix[graph_index],
                                                    activation=tf.nn.tanh)

                    f_k_g = tf.reshape(f_k_g, [-1, self._T, station_number, self._input_dim])

                    x_hat = tf.concat([f_k_g, traffic_flow], axis=-1)

                    z = tf.reduce_mean(x_hat, axis=-2, keepdims=True)

                    s = tf.layers.dense(tf.layers.dense(z, units=4, use_bias=False, activation=tf.nn.relu),
                                        units=1, use_bias=False, activation=tf.nn.sigmoid)

                    x_rnn = tf.multiply(traffic_flow, tf.tile(s, [1, 1, station_number, self._input_dim]))

                    x_rnn = tf.reshape(tf.transpose(x_rnn, perm=[0, 2, 1, 3]), [-1, self._T, self._input_dim])

                    for lstm_layer_index in range(self._lstm_layers):
                        x_rnn = tf.keras.layers.LSTM(units=self._lstm_units,
                                                     activation='tanh',
                                                     dropout=0.1,
                                                     kernel_regularizer=tf.contrib.layers.l2_regularizer(1e-4),
                                                     return_sequences=True if lstm_layer_index<self._lstm_layers-1
                                                                      else False)\
                                                    (x_rnn)

                    H = tf.reshape(x_rnn, [-1, station_number, self._lstm_units])

                    outputs.append(H)

            outputs = tf.reduce_sum(outputs, axis=0)

            # external dims
            if self._external_dim is not None and self._external_dim > 0:
                external_input = tf.placeholder(tf.float32, [None, self._external_dim])
                self._input['external_feature'] = external_input.name
                external_dense = tf.layers.dense(inputs=external_input, units=10)
                external_dense = tf.tile(tf.reshape(external_dense, [-1, 1, 10]),
                                         [1, tf.shape(outputs)[-2], 1])
                outputs = tf.concat([outputs, external_dense], axis=-1)

            prediction = tf.layers.dense(outputs, units=1)

            loss = tf.sqrt(tf.reduce_mean(tf.square(prediction - target)))

            train_operation = tf.train.AdamOptimizer(self._lr).minimize(loss, name='train_op')

            # record output
            self._output['prediction'] = prediction.name
            self._output['loss'] = loss.name

            # record train operation
            self._op['train_op'] = train_operation.name

            super(ST_MGCN, self).build(init_vars=init_vars, max_to_keep=max_to_keep)

    # Step 1 : Define your '_get_feed_dict function‘, map your input to the tf-model
    def _get_feed_dict(self, traffic_flow, laplace_matrix, target=None, external_feature=None):
        feed_dict = {
            'traffic_flow': traffic_flow,
            'laplace_matrix': laplace_matrix,
        }
        if target is not None:
            feed_dict['target'] = target
        if external_feature is not None:
            feed_dict['external_feature'] = external_feature
        return feed_dict



import torch
import torch.nn as nn
import numpy as np
import os
import logging
import math
import time
import copy
from UCTB.preprocess import SplitData
from UCTB.train.LossFunction import masked_mae_loss


class Trainer(object):
    def __init__(self, model, train_loader, val_loader, test_loader,
                 args):
        super(Trainer, self).__init__()
        self.model = model
        self.optimizer = torch.optim.Adam(params=model.parameters(), lr=args.lr_init, eps=1.0e-8,
                                          weight_decay=0, amsgrad=False)
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.test_loader = test_loader
        self.args = args
        self.train_per_epoch = len(train_loader)
        self.lr_scheduler = None
        for p in self.model.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)
            else:
                nn.init.uniform_(p)
        if args.loss_func == 'mask_mae':
            self.loss = masked_mae_loss(mask_value=0.0)
        elif args.loss_func == 'mae':
            self.loss = torch.nn.L1Loss().to(args.device)
        elif args.loss_func == 'mse':
            self.loss = torch.nn.MSELoss().to(args.device)
        else:
            raise ValueError
        if args.lr_decay:
            print('Applying learning rate decay.')
            lr_decay_steps = [int(i)
                              for i in list(args.lr_decay_step.split(','))]
            self.lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer=self.optimizer,
                                                                     milestones=lr_decay_steps,
                                                                     gamma=args.lr_decay_rate)
        if val_loader != None:
            self.val_per_epoch = len(val_loader)
        self.best_path = os.path.join(self.args.log_dir, 'best_model.pth')
        self.loss_figure_path = os.path.join(self.args.log_dir, 'loss.png')
        # log
        if os.path.isdir(args.log_dir) == False and not args.debug:
            os.makedirs(args.log_dir, exist_ok=True)
        self.logger = get_logger(
            args.log_dir, name=args.model, debug=args.debug)
        self.logger.info('Experiment log path in: {}'.format(args.log_dir))
        # if not args.debug:
        # self.logger.info("Argument: %r", args)
        # for arg, value in sorted(vars(args).items()):
        #     self.logger.info("Argument %s: %r", arg, value)

    def val_epoch(self, epoch, val_dataloader):
        self.model.eval()
        total_val_loss = 0

        with torch.no_grad():
            for batch_idx, (data, target) in enumerate(val_dataloader):
                label = target[..., :self.args.output_dim]
                output = self.model(data, target, teacher_forcing_ratio=0.)
                #loss = self.loss(output.cuda(), label)
                loss = self.loss(output, label)
                # a whole batch of Metr_LA is filtered
                if not torch.isnan(loss):
                    total_val_loss += loss.item()
        val_loss = total_val_loss / len(val_dataloader)
        self.logger.info(
            '**********Val Epoch {}: average Loss: {:.6f}'.format(epoch, val_loss))
        return val_loss

    def train_epoch(self, epoch):
        self.model.train()
        total_loss = 0
        for batch_idx, (data, target) in enumerate(self.train_loader):
            label = target[..., :self.args.output_dim]  # (..., 1)
            self.optimizer.zero_grad()

            # teacher_forcing for RNN encoder-decoder model
            # if teacher_forcing_ratio = 1: use label as input in the decoder for all steps
            if self.args.teacher_forcing:
                global_step = (epoch - 1) * self.train_per_epoch + batch_idx
                teacher_forcing_ratio = self._compute_sampling_threshold(
                    global_step, self.args.tf_decay_steps)
            else:
                teacher_forcing_ratio = 1.
            # data and target shape: B, T, N, F; output shape: B, T, N, F
            output = self.model(
                data, target, teacher_forcing_ratio=teacher_forcing_ratio)
            #loss = self.loss(output.cuda(), label)
            loss = self.loss(output, label)
            loss.backward()

            # add max grad clipping
            if self.args.grad_norm:
                torch.nn.utils.clip_grad_norm_(
                    self.model.parameters(), self.args.max_grad_norm)
            self.optimizer.step()
            total_loss += loss.item()

            # log information
            if batch_idx % self.args.log_step == 0:
                self.logger.info('Train Epoch {}: {}/{} Loss: {:.6f}'.format(
                    epoch, batch_idx, self.train_per_epoch, loss.item()))
        train_epoch_loss = total_loss / self.train_per_epoch
        self.logger.info(
            '**********Train Epoch {}: averaged Loss: {:.6f}, tf_ratio: {:.6f}'.format(epoch, train_epoch_loss,
                                                                                       teacher_forcing_ratio))

        # learning rate decay
        if self.args.lr_decay:
            self.lr_scheduler.step()
        return train_epoch_loss

    def train(self):
        best_model = None
        best_loss = float('inf')
        not_improved_count = 0
        train_loss_list = []
        val_loss_list = []
        start_time = time.time()
        for epoch in range(1, self.args.epochs + 1):
            # epoch_time = time.time()
            train_epoch_loss = self.train_epoch(epoch)
            # print(time.time()-epoch_time)
            # exit()
            if self.val_loader == None:
                val_dataloader = self.test_loader
            else:
                val_dataloader = self.val_loader
            val_epoch_loss = self.val_epoch(epoch, val_dataloader)

            # print('LR:', self.optimizer.param_groups[0]['lr'])
            train_loss_list.append(train_epoch_loss)
            val_loss_list.append(val_epoch_loss)
            if train_epoch_loss > 1e6:
                self.logger.warning('Gradient explosion detected. Ending...')
                break
            # if self.val_loader == None:
            # val_epoch_loss = train_epoch_loss
            if val_epoch_loss < best_loss:
                best_loss = val_epoch_loss
                not_improved_count = 0
                best_state = True
            else:
                not_improved_count += 1
                best_state = False
            # early stop
            if self.args.early_stop:
                if not_improved_count == self.args.early_stop_patience:
                    self.logger.info("Validation performance didn\'t improve for {} epochs. "
                                     "Training stops.".format(self.args.early_stop_patience))
                    break
            # save the best state
            if best_state == True:
                self.logger.info(
                    '*********************************Current best model saved!')
                best_model = copy.deepcopy(self.model.state_dict())

        training_time = time.time() - start_time
        self.logger.info("Total training time: {:.4f}min, best loss: {:.6f}".format(
            (training_time / 60), best_loss))

        # save the best model to file
        if not self.args.debug:
            torch.save(best_model, self.best_path)
            self.logger.info("Saving current best model to " + self.best_path)

        # test
        self.model.load_state_dict(best_model)
        # self.val_epoch(self.args.epochs, self.test_loader)
        #self.test(self.model, self.args, self.test_loader, self.scaler, self.logger)

    def save_checkpoint(self):
        state = {
            'state_dict': self.model.state_dict(),
            'optimizer': self.optimizer.state_dict(),
            'config': self.args
        }
        torch.save(state, self.best_path)
        self.logger.info("Saving current best model to " + self.best_path)

    @staticmethod
    def test(model, args, data_loader,  logger, path=None):
        if path != None:
            check_point = torch.load(path)
            state_dict = check_point['state_dict']
            args = check_point['config']
            model.load_state_dict(state_dict)
            model.to(args.device)
        model.eval()
        y_pred = []
        y_true = []
        with torch.no_grad():
            for batch_idx, (data, target) in enumerate(data_loader):
                label = target[..., :args.output_dim]
                output = model(data, target, teacher_forcing_ratio=0)
                y_true.append(label)
                y_pred.append(output)
        y_pred = torch.cat(y_pred, dim=0)
        y_pred = y_pred.cpu().numpy()

        return y_pred

    @staticmethod
    def _compute_sampling_threshold(global_step, k):
        """
        Computes the sampling probability for scheduled sampling using inverse sigmoid.
        :param global_step:
        :param k:
        :return:
        """
        return k / (k + math.exp(global_step / k))


def get_dataloader_AGCRN(data_loader, batchsize, tod=False, dow=False, weather=False, single=True):
    # split data
    train_closeness, val_closeness = SplitData.split_data(
        data_loader.train_closeness, [0.9, 0.1])
    train_period, val_period = SplitData.split_data(
        data_loader.train_period, [0.9, 0.1])
    train_trend, val_trend = SplitData.split_data(
        data_loader.train_trend, [0.9, 0.1])
    train_y, val_y = SplitData.split_data(data_loader.train_y, [0.9, 0.1])

    # T, N, D, 1 -> T, D, N, 1
    if data_loader.period_len > 0 and data_loader.trend_len > 0:
        train_x = np.concatenate(
            [train_trend, train_period, train_closeness], axis=2).transpose([0, 2, 1, 3])
        val_x = np.concatenate(
            [val_trend, val_period, val_closeness], axis=2).transpose([0, 2, 1, 3])
        test_x = np.concatenate([data_loader.test_trend, data_loader.test_period, data_loader.test_closeness],
                                axis=2).transpose([0, 2, 1, 3])
    else:
        train_x = train_closeness.transpose([0, 2, 1, 3])
        val_x = val_closeness.transpose([0, 2, 1, 3])
        test_x = data_loader.test_closeness.transpose([0, 2, 1, 3])

    train_y = train_y[:, np.newaxis]
    val_y = val_y[:, np.newaxis]
    test_y = data_loader.test_y[:, np.newaxis]

    print('Train: ', train_x.shape, train_y.shape)
    print('Val: ', val_x.shape, val_y.shape)
    print('Test: ', test_x.shape, test_y.shape)
    ############## get dataloader ######################
    train_dataloader = data_loader_torch(
        train_x, train_y, batchsize, shuffle=True, drop_last=True)
    if len(train_x) == 0:
        val_dataloader = None
    else:
        val_dataloader = data_loader_torch(
            val_x, val_y, batchsize, shuffle=False, drop_last=True)
    test_dataloader = data_loader_torch(
        test_x, test_y, batchsize, shuffle=False, drop_last=False)
    return train_dataloader, val_dataloader, test_dataloader


def data_loader_torch(X, Y, batch_size, shuffle=True, drop_last=True):
    cuda = True if torch.cuda.is_available() else False
    TensorFloat = torch.cuda.FloatTensor if cuda else torch.FloatTensor
    X, Y = TensorFloat(X), TensorFloat(Y)
    data = torch.utils.data.TensorDataset(X, Y)
    dataloader = torch.utils.data.DataLoader(data, batch_size=batch_size,
                                             shuffle=shuffle, drop_last=drop_last)
    return dataloader


def get_logger(root, name=None, debug=True):
    #when debug is true, show DEBUG and INFO in screen
    #when debug is false, show DEBUG in file and info in both screen&file
    #INFO will always be in screen
    # create a logger
    logger = logging.getLogger(name)
    #critical > error > warning > info > debug > notset
    logger.setLevel(logging.DEBUG)

    # define the formate
    formatter = logging.Formatter('%(asctime)s: %(message)s', "%Y-%m-%d %H:%M")
    # create another handler for output log to console
    console_handler = logging.StreamHandler()
    if debug:
        console_handler.setLevel(logging.DEBUG)
    else:
        console_handler.setLevel(logging.INFO)
        # create a handler for write log to file
        logfile = os.path.join(root, 'run.log')
        print('Creat Log File in: ', logfile)
        file_handler = logging.FileHandler(logfile, mode='w')
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)
    # add Handler to logger
    logger.addHandler(console_handler)
    if not debug:
        logger.addHandler(file_handler)
    return logger



import numpy as np
from abc import ABC, abstractmethod


class Normalizer(ABC):
    """
    Normalizer is the base abstract class for many normalizers such as MaxMinNormalizer and ZscoreNormalizer.You can also build your own normalizer by inheriting this class.

    Args:
        X(ndarray): Data which normalizer extracts characteristics from.
    """
    @abstractmethod
    def __init__(self, X):
        pass
    @abstractmethod
    def transform(self, X_in):
        pass
    @abstractmethod
    def inverse_transform(self, X_in):
        pass



class MaxMinNormalizer(Normalizer):
    '''
    This class can help normalize and denormalize data using maximum and minimum of data by calling transform and inverse_transform method.

    Args:
        X(ndarray): Data which normalizer extracts characteristics from.
        method(str): Parameter to choose in which way the input data will be processed.
    '''
    def __init__(self, X,method='all'):
        self.method = method
        self._min = np.min(X)
        self._max = np.max(X)
        self._min_by_column = np.min(X,axis=0)
        self._max_by_column = np.max(X,axis=0)

    def transform(self, X):
        '''
        Process input data to obtain normalized data.

        Args:
            X(ndarray): input data. 
        :return: normalized data.
        :type: numpy.ndarray.
        '''
        if self.method=='all':
            return (X - self._min) / (self._max - self._min)
        elif self.method=='column':
            return (X - self._min_by_column) / (self._max_by_column - self._min_by_column)

    def inverse_transform(self, X):
        '''
        Restore normalized data.

        Args:
            X(ndarray): normalized data. 
        :return: denormalized data.
        :type: numpy.ndarray.
        '''
        if self.method=='all':
            return X * (self._max - self._min) + self._min
        elif self.method=='column':
            return X * (self._max_by_column - self._min_by_column) + self._min_by_column


class WhiteNormalizer(Normalizer):
    '''
    This class's normalization won't do anything.
    '''
    def __init__(self, X,method='all'):
        pass

    def transform(self, X):
        '''
        Process input data to obtain normalized data.

        Args:
            X(ndarray): input data. 
        :return: normalized data.
        :type: numpy.ndarray.
        '''
        return X

    def inverse_transform(self, X):
        '''
        Restore normalized data.

        Args:
            X(ndarray): normalized data. 
        :return: denormalized data.
        :type: numpy.ndarray.
        '''
        return X

class ZscoreNormalizer(Normalizer):
    '''
    This class can help normalize and denormalize data using mean and standard deviation in data by calling transform and inverse_transform method.

    Args:
        X(ndarray): Data which normalizer extracts characteristics from.
        method(str): Parameter to choose in which way the input data will be processed.
    '''
    def __init__(self, X,method='all'):
        self.method = method
        self._mean = np.mean(X)
        self._std = np.std(X)
        self._mean_by_column = np.mean(X,axis=0)
        self._std_by_column = np.std(X,axis=0)

    def transform(self, X):
        '''
        Process input data to obtain normalized data.

        Args:
            X(ndarray): input data. 
        :return: normalized data.
        :type: numpy.ndarray.
        '''
        if self.method=='all':
            return (X - self._mean) / self._std
        elif self.method=='column':
            return (X - self._mean_by_column) / self._std_by_column

    def inverse_transform(self, X):
        '''
        Restore normalized data.

        Args:
            X(ndarray): normalized data. 
        :return: denormalized data.
        :type: numpy.ndarray.
        '''
        if self.method=='all':
            return X * self._std + self._mean
        elif self.method=='column':
            return X * self._std_by_column + self._mean_by_column


class MoveSample(object):
    def __init__(self, feature_step, feature_stride, feature_length, target_length):
        self.feature_step = feature_step
        self.feature_stride = feature_stride
        self.feature_length = feature_length
        self.target_length = target_length

    def general_move_sample(self, data):
        feature = []
        target = []
        for i in range(len(data) - self.feature_length -
                       (self.feature_step-1)*self.feature_stride - self.target_length + 1):
            feature.append([data[i + step*self.feature_stride: i + step*self.feature_stride + self.feature_length]
                            for step in range(self.feature_step)])
            target.append(data[i + (self.feature_step-1) * self.feature_stride + self.feature_length:\
                               i + (self.feature_step-1) * self.feature_stride + self.feature_length + self.target_length])

        return np.array(feature), np.array(target)

class ST_MoveSample(object):
    '''
    This class can converts raw data into temporal features including closenss, period and trend features.

    Args:
        closeness_len(int):The length of closeness data history. The former consecutive ``closeness_len`` time slots
            of data will be used as closeness history.
        period_len(int):The length of period data history. The data of exact same time slots in former consecutive
            ``period_len`` days will be used as period history.
        trend_len(int):The length of trend data history. The data of exact same time slots in former consecutive
            ``trend_len`` weeks (every seven days) will be used as trend history.
        target_length(int):The numbers of steps that need prediction by one piece of history data. Have to be 1 now.
            Default: 1 default:1.
        daily_slots(int): The number of records of one day. Calculated by 24 * 60 /time_fitness. default:24.
    '''
    def __init__(self, closeness_len, period_len, trend_len, target_length=1, daily_slots=24):
        self._c_t = closeness_len
        self._p_t = period_len
        self._t_t = trend_len
        self._target_length = target_length
        self._daily_slots = daily_slots

        # 1 init Move_Sample object
        self.move_sample_closeness = MoveSample(feature_step=self._c_t, feature_stride=1,
                                                feature_length=1, target_length=self._target_length)

        self.move_sample_period = MoveSample(feature_step=self._p_t + 1, feature_stride=int(self._daily_slots),
                                             feature_length=1, target_length=0)

        self.move_sample_trend = MoveSample(feature_step=self._t_t + 1, feature_stride=int(self._daily_slots) * 7,
                                            feature_length=1, target_length=0)

    def move_sample(self, data):
        '''
        Input data to generate closeness, period, trend features and target vector y.

        Args:
            data(ndarray):Orginal temporal data.
        :return:closeness, period, trend and y matrices.
        :type: numpy.ndarray.
        '''
        # 2 general move sample
        closeness, y = self.move_sample_closeness.general_move_sample(data)
        period, _ = self.move_sample_period.general_move_sample(data)
        trend, _ = self.move_sample_trend.general_move_sample(data)

        # 3 remove the front part
        min_length = min(len(closeness), len(period), len(trend))
        closeness = closeness[-min_length:]
        y = y[-min_length:]
        period = period[-min_length:]
        trend = trend[-min_length:]

        # 4 remove tail of period and trend
        period = period[:, :-1]
        trend = trend[:, :-1]

        if self._c_t and self._c_t > 0:
            closeness = np.transpose(closeness, [0] + list(range(3, len(closeness.shape))) + [1, 2])
        else:
            closeness = np.array([])

        if self._p_t and self._p_t > 0:
            period = np.transpose(period, [0] + list(range(3, len(period.shape))) + [1, 2])
        else:
            period = np.zeros(shape=[min_length,closeness.shape[1],0,1])

        if self._t_t and self._t_t > 0:
            trend = np.transpose(trend, [0] + list(range(3, len(trend.shape))) + [1, 2])
        else:
            trend = np.zeros(shape=[min_length,closeness.shape[1],0,1])

        y = np.transpose(y, [0] + list(range(2, len(y.shape))) + [1])

        return closeness, period, trend, y


class SplitData(object):
    '''
    This class can help split data by calling split_data and split_feed_dict method.
    '''
    @staticmethod
    def split_data(data, ratio_list):
        '''
        Divide the data based on the given parameter ratio_list.
        
        Args:
            data(ndarray):Data to be split.
            ratio_list(list):Split ratio, the `data` will be split according to the ratio.
        :return:The elements in the returned list are the divided data, and the 
            dimensions of the list are the same as ratio_list.
        :type: list
        '''
        if np.sum(ratio_list) != 1:
            ratio_list = np.array(ratio_list)
            ratio_list = ratio_list / np.sum(ratio_list)
        return [data[int(sum(ratio_list[0:e])*len(data)):
                     int(sum(ratio_list[0:e+1])*len(data))] for e in range(len(ratio_list))]

    @staticmethod
    def split_feed_dict(feed_dict, sequence_length, ratio_list):
        '''
        Divide the `value` data in `feed_dict` based on the given parameter ratio_list.

        Args:
            feed_dict(dict):It is a dictionary composed of `key-value` pairs.
            sequence_length(int):If the length of `value` in `feed_dict` is equal to sequence_length, 
                then this method divides the `value` according to the ratio without changing its `key`.
            ratio_list(list):Split ratio, the data will be split according to the ratio.
        :return: The elements in the returned list are divided dictionaries, and the dimensions of the list are the same as ratio_list.
        :type: list
        '''
        if np.sum(ratio_list) != 1:
            ratio_list = np.array(ratio_list)
            ratio_list = ratio_list / np.sum(ratio_list)

        return [{key: value[int(sum(ratio_list[0:e])*len(value)):int(sum(ratio_list[0:e+1])*len(value))]
                 if len(value) == sequence_length else value for key, value in feed_dict.items()}
                for e in range(len(ratio_list))]

def chooseNormalizer(in_arg,X_train):
    '''
    Choose a proper normalizer consistent with user's input.

    Args:
        in_arg(str|bool|object):Function is based on it to choose different normalizer.
        X_train(numpy.ndarray):Function is based on it to initialize the normalizer.
    :return: The normalizer consistent with definition.
    :type: object.
    '''
    if type(in_arg) == str:
        if '-' in in_arg:
            method,way=in_arg.split('-')
            if method=='Zscore' or method=='zscore' or method=='ZScore':
                return ZscoreNormalizer(X_train,way)
            elif method=='MaxMin' or method=='maxmin' or method=='Maxmin' or method=='MinMax' or method=='Minmax' or method=='minmax':
                return MaxMinNormalizer(X_train,way)
            else:
                raise ValueError('We havn\'t support thie method for normalization yet')
        else:
            raise ValueError('We don\'t accept this format of str input for how to do normalization')
    elif type(in_arg) == bool:
        if in_arg:
            return MaxMinNormalizer(X_train)
        else:
            return WhiteNormalizer(X_train)
    elif type(in_arg) == object:
        if hasattr(in_arg,'transform') and hasattr(in_arg,'inverss_transform'):
            return in_arg(X_train)
        else:
            raise TypeError('Your custom normalizer is not in compliance')
    else:
        raise TypeError('We don\'t accept {} of input for how to do normalization')




import os
import copy
import datetime
import numpy as np
from dateutil.parser import parse
from sklearn.metrics.pairwise import cosine_similarity
from scipy.stats import pearsonr
from ..preprocess.time_utils import is_work_day_china, is_work_day_america, is_valid_date
from ..preprocess import MoveSample, SplitData, ST_MoveSample, chooseNormalizer
from .dataset import DataSet
from ..preprocess.preprocessor import *

class GridTrafficLoader(object):

    def __init__(self,
                 dataset,
                 city=None,
                 data_range='all',
                 train_data_length='all',
                 test_ratio=0.1,
                 closeness_len=6,
                 period_len=7,
                 trend_len=4,
                 target_length=1,
                 normalize=True,
                 workday_parser=is_work_day_america,
                 data_dir=None,
                 MergeIndex=1,
                 MergeWay="sum",**kwargs):
                 
        self.dataset = DataSet(dataset, MergeIndex, MergeWay, city,data_dir=data_dir)

        self.loader_id = "{}_{}_{}_{}_{}_{}_{}_G".format(data_range, train_data_length, test_ratio, closeness_len, period_len, trend_len, self.dataset.time_fitness)

        self.daily_slots = 24 * 60 / self.dataset.time_fitness

        if type(data_range) is str and data_range.lower() == 'all':
            data_range = [0, len(self.dataset.grid_traffic)]
        elif type(data_range) is float:
            data_range = [0, int(data_range * len(self.dataset.grid_traffic))]
        else:
            data_range = [int(data_range[0] * self.daily_slots), int(data_range[1] * self.daily_slots)]

        num_time_slots = data_range[1] - data_range[0]
        self.traffic_data = self.dataset.grid_traffic[data_range[0]:data_range[1], :].astype(np.float32)

        # external feature
        external_feature = []
        # weather
        if len(self.dataset.external_feature_weather) > 0:
            external_feature.append(self.dataset.external_feature_weather[data_range[0]:data_range[1]])
            # Weekday Feature
            weekday_feature = [[1 if workday_parser(parse(self.dataset.time_range[0])
                                                    + datetime.timedelta(hours=e * self.dataset.time_fitness / 60), self.dataset.city) else 0] \
                            for e in range(data_range[0], num_time_slots + data_range[0])]
            # Hour Feature
            hour_feature = [[(parse(self.dataset.time_range[0]) +
                            datetime.timedelta(hours=e * self.dataset.time_fitness / 60)).hour / 24.0]
                            for e in range(data_range[0], num_time_slots + data_range[0])]

            external_feature.append(weekday_feature)
            external_feature.append(hour_feature)
            external_feature = np.concatenate(external_feature, axis=-1).astype(np.float32)
            self.external_dim = external_feature.shape[1]
        else:
            self.external_dim = len(external_feature)

        self.height, self.width = self.traffic_data.shape[1], self.traffic_data.shape[2]

        if test_ratio > 1 or test_ratio < 0:
            raise ValueError('test_ratio ')
        train_test_ratio = [1 - test_ratio, test_ratio]

        self.train_data, self.test_data = SplitData.split_data(self.traffic_data, train_test_ratio)
        self.train_ef, self.test_ef = SplitData.split_data(external_feature, train_test_ratio)

        # Normalize the traffic data
        if normalize:
            self.normalizer = Normalizer(self.train_data)
            self.train_data = self.normalizer.min_max_normal(self.train_data)
            self.test_data = self.normalizer.min_max_normal(self.test_data)

        if train_data_length.lower() != 'all':
            train_day_length = int(train_data_length)
            self.train_data = self.train_data[-int(train_day_length * self.daily_slots):]
            self.train_ef = self.train_ef[-int(train_day_length * self.daily_slots):]

        # expand the test data
        expand_start_index = len(self.train_data) - max(int(self.daily_slots * period_len),
                                                        int(self.daily_slots * 7 * trend_len), closeness_len)

        self.test_data = np.vstack([self.train_data[expand_start_index:], self.test_data])
        self.test_ef = np.vstack([self.train_ef[expand_start_index:], self.test_ef])

        assert type(closeness_len) is int and closeness_len >= 0
        assert type(period_len) is int and period_len >= 0
        assert type(trend_len) is int and trend_len >= 0

        self.closeness_len = closeness_len
        self.period_len = period_len
        self.trend_len = trend_len

        # init move sample obj
        self.st_move_sample = ST_MoveSample(closeness_len=closeness_len,
                                            period_len=period_len,
                                            trend_len=trend_len, target_length=target_length, daily_slots=self.daily_slots)

        self.train_closeness, \
        self.train_period, \
        self.train_trend, \
        self.train_y = self.st_move_sample.move_sample(self.train_data)

        self.test_closeness, \
        self.test_period, \
        self.test_trend, \
        self.test_y = self.st_move_sample.move_sample(self.test_data)

        self.train_closeness = self.train_closeness.squeeze(-1)
        self.train_period = self.train_period.squeeze(-1)
        self.train_trend = self.train_trend.squeeze(-1)

        self.test_closeness = self.test_closeness.squeeze(-1)
        self.test_period = self.test_period.squeeze(-1)
        self.test_trend = self.test_trend.squeeze(-1)

        self.train_sequence_len = max((len(self.train_closeness), len(self.train_period), len(self.train_trend)))
        self.test_sequence_len = max((len(self.test_closeness), len(self.test_period), len(self.test_trend)))

        # external feature
        self.train_ef = self.train_ef[-self.train_sequence_len - target_length: -target_length]
        self.test_ef = self.test_ef[-self.test_sequence_len - target_length: -target_length]


class NodeTrafficLoader(object):
    """The data loader that extracts and processes data from a :obj:`DataSet` object.

    Args:
        dataset (str): A string containing path of the dataset pickle file or a string of name of the dataset.
        city (:obj:`str` or ``None``): ``None`` if dataset is file path, or a string of name of the city.
            Default: ``None``
        data_range: The range of data extracted from ``self.dataset`` to be further used. If set to ``'all'``, all data in
            ``self.dataset`` will be used. If set to a float between 0.0 and 1.0, the relative former proportion of data in
            ``self.dataset`` will be used. If set to a list of two integers ``[start, end]``, the data from *start* day to
            (*end* - 1) day of data in ``self.dataset`` will be used. Default: ``'all'``
        train_data_length: The length of train data. If set to ``'all'``, all data in the split train set will be used.
            If set to int, the latest ``train_data_length`` days of data will be used as train set. Default: ``'all'``
        test_ratio (float): The ratio of test set as data will be split into train set and test set. Default: 0.1
        closeness_len (int): The length of closeness data history. The former consecutive ``closeness_len`` time slots
            of data will be used as closeness history. Default: 6
        period_len (int): The length of period data history. The data of exact same time slots in former consecutive
            ``period_len`` days will be used as period history. Default: 7
        trend_len (int): The length of trend data history. The data of exact same time slots in former consecutive
            ``trend_len`` weeks (every seven days) will be used as trend history. Default: 4
        target_length (int): The numbers of steps that need prediction by one piece of history data. Have to be 1 now.
            Default: 1
        normalize (bool|str|object): Select which normalizer to normalize input data. Default: ``True``
        workday_parser: Used to build external features to be used in neural methods. Default: ``is_work_day_america``
        with_tpe (bool): If ``True``, data loader will build time position embeddings. Default: ``False``
        data_dir (:obj:`str` or ``None``): The dataset directory. If set to ``None``, a directory will be created. If
            ``dataset`` is file path, ``data_dir`` should be ``None`` too. Default: ``None``
        MergeIndex(int): The granularity of dataset will be ``MergeIndex`` * original granularity.
        MergeWay(str): How to change the data granularity. Now it can be ``sum`` ``average`` or ``max``.
        remove(bool): If ``True``, dataloader  will remove stations whose average traffic is less than 1. 
            Othewise, dataloader will use all stations.

    Attributes:
        dataset (DataSet): The DataSet object storing basic data.
        daily_slots (int): The number of time slots in one single day.
        station_number (int): The number of nodes.
        external_dim (int): The number of dimensions of external features.
        train_closeness (np.ndarray): The closeness history of train set data. When ``with_tpe`` is ``False``,
            its shape is [train_time_slot_num, ``station_number``, ``closeness_len``, 1].
            On the dimension of ``closeness_len``, data are arranged from earlier time slots to later time slots.
            If ``closeness_len`` is set to 0, train_closeness will be an empty ndarray.
            ``train_period``, ``train_trend``, ``test_closeness``, ``test_period``, ``test_trend`` have similar shape
            and construction.
        train_y (np.ndarray): The train set data. Its shape is [train_time_slot_num, ``station_number``, 1].
            ``test_y`` has similar shape and construction.
    """

    def __init__(self,
                 dataset,
                 city=None,
                 data_range='all',
                 train_data_length='all',
                 test_ratio=0.1,
                 closeness_len=6,
                 period_len=7,
                 trend_len=4,
                 target_length=1,
                 normalize=True,
                 workday_parser=is_work_day_america,
                 with_tpe=False,
                 data_dir=None,
                 MergeIndex=1,
                 MergeWay="sum",
                 remove=True,**kwargs):

        self.dataset = DataSet(dataset, MergeIndex, MergeWay, city,data_dir=data_dir)

        self.loader_id = "{}_{}_{}_{}_{}_{}_{}_N".format(data_range, train_data_length, test_ratio, closeness_len, period_len, trend_len, self.dataset.time_fitness)
        
        self.daily_slots = int(24 * 60 / self.dataset.time_fitness)

        self.closeness_len = int(closeness_len)
        self.period_len = int(period_len)
        self.trend_len = int(trend_len)

        assert type(self.closeness_len) is int and self.closeness_len >= 0
        assert type(self.period_len) is int and self.period_len >= 0
        assert type(self.trend_len) is int and self.trend_len >= 0

        if type(data_range) is str and data_range.lower().startswith("0."):
            data_range = float(data_range)
        if type(data_range) is str and data_range.lower() == 'all':
            data_range = [0, len(self.dataset.node_traffic)]
        elif type(data_range) is float:
            data_range = [0, int(data_range * len(self.dataset.node_traffic))]
        else:
            data_range = [int(data_range[0] * self.daily_slots), int(data_range[1] * self.daily_slots)]

        num_time_slots = data_range[1] - data_range[0]

        # traffic feature
        if remove:
            self.traffic_data_index = np.where(np.mean(self.dataset.node_traffic, axis=0) * self.daily_slots > 1)[0]
        else:
            self.traffic_data_index = np.arange(self.dataset.node_traffic.shape[1])

        self.traffic_data = self.dataset.node_traffic[data_range[0]:data_range[1], self.traffic_data_index].astype(
             np.float32)

        # external feature
        external_feature = []
        # weather
        if len(self.dataset.external_feature_weather) > 0:
            external_feature.append(self.dataset.external_feature_weather[data_range[0]:data_range[1]])
            # Weekday Feature
            weekday_feature = [[1 if workday_parser(parse(self.dataset.time_range[0])
                                                    + datetime.timedelta(hours=e * self.dataset.time_fitness / 60), self.dataset.city) else 0] \
                            for e in range(data_range[0], num_time_slots + data_range[0])]
            # Hour Feature
            hour_feature = [[(parse(self.dataset.time_range[0]) +
                            datetime.timedelta(hours=e * self.dataset.time_fitness / 60)).hour / 24.0]
                            for e in range(data_range[0], num_time_slots + data_range[0])]

            external_feature.append(weekday_feature)
            external_feature.append(hour_feature)
            external_feature = np.concatenate(external_feature, axis=-1).astype(np.float32)
            self.external_dim = external_feature.shape[1]
        else:
            self.external_dim = len(external_feature)
        
        self.station_number = self.traffic_data.shape[1]
        
        if test_ratio > 1 or test_ratio < 0:
            raise ValueError('test_ratio ')
        self.train_test_ratio = [1 - test_ratio, test_ratio]

        self.train_data, self.test_data = SplitData.split_data(self.traffic_data, self.train_test_ratio)
        self.train_ef, self.test_ef = SplitData.split_data(external_feature, self.train_test_ratio)

        # Normalize the traffic data

        self.normalizer = chooseNormalizer(normalize,self.train_data)
        self.train_data = self.normalizer.transform(self.train_data)
        self.test_data = self.normalizer.transform(self.test_data)

        if train_data_length.lower() != 'all':
            train_day_length = int(train_data_length)
            self.train_data = self.train_data[-int(train_day_length * self.daily_slots):]
            self.train_ef = self.train_ef[-int(train_day_length * self.daily_slots):]

        # expand the test data
        expand_start_index = len(self.train_data) - \
                             max(int(self.daily_slots * self.period_len),
                                 int(self.daily_slots * 7 * self.trend_len),
                                 self.closeness_len)

        self.test_data = np.vstack([self.train_data[expand_start_index:], self.test_data])
        self.test_ef = np.vstack([self.train_ef[expand_start_index:], self.test_ef])

        # init move sample obj
        self.st_move_sample = ST_MoveSample(closeness_len=self.closeness_len,
                                            period_len=self.period_len,
                                            trend_len=self.trend_len, target_length=target_length, daily_slots=self.daily_slots)

        self.train_closeness, \
        self.train_period, \
        self.train_trend, \
        self.train_y = self.st_move_sample.move_sample(self.train_data)

        self.test_closeness, \
        self.test_period, \
        self.test_trend, \
        self.test_y = self.st_move_sample.move_sample(self.test_data)

        self.train_sequence_len = max((len(self.train_closeness), len(self.train_period), len(self.train_trend)))
        self.test_sequence_len = max((len(self.test_closeness), len(self.test_period), len(self.test_trend)))

        # external feature
        self.train_ef = self.train_ef[-self.train_sequence_len - target_length: -target_length]
        self.test_ef = self.test_ef[-self.test_sequence_len - target_length: -target_length]

        if with_tpe:

            # Time position embedding
            self.closeness_tpe = np.array(range(1, self.closeness_len + 1), dtype=np.float32)
            self.period_tpe = np.array(range(1 * int(self.daily_slots),
                                             self.period_len * int(self.daily_slots) + 1,
                                             int(self.daily_slots)), dtype=np.float32)
            self.trend_tpe = np.array(range(1 * int(self.daily_slots) * 7,
                                            self.trend_len * int(self.daily_slots) * 7 + 1,
                                            int(self.daily_slots) * 7), dtype=np.float32)

            self.train_closeness_tpe = np.tile(np.reshape(self.closeness_tpe, [1, 1, -1, 1]),
                                               [len(self.train_closeness), len(self.traffic_data_index), 1, 1])
            self.train_period_tpe = np.tile(np.reshape(self.period_tpe, [1, 1, -1, 1]),
                                            [len(self.train_period), len(self.traffic_data_index), 1, 1])
            self.train_trend_tpe = np.tile(np.reshape(self.trend_tpe, [1, 1, -1, 1]),
                                           [len(self.train_trend), len(self.traffic_data_index), 1, 1])

            self.test_closeness_tpe = np.tile(np.reshape(self.closeness_tpe, [1, 1, -1, 1]),
                                              [len(self.test_closeness), len(self.traffic_data_index), 1, 1])
            self.test_period_tpe = np.tile(np.reshape(self.period_tpe, [1, 1, -1, 1]),
                                           [len(self.test_period), len(self.traffic_data_index), 1, 1])
            self.test_trend_tpe = np.tile(np.reshape(self.trend_tpe, [1, 1, -1, 1]),
                                          [len(self.test_trend), len(self.traffic_data_index), 1, 1])

            self.tpe_dim = self.train_closeness_tpe.shape[-1]

            # concat temporal feature with time position embedding
            self.train_closeness = np.concatenate((self.train_closeness, self.train_closeness_tpe,), axis=-1)
            self.train_period = np.concatenate((self.train_period, self.train_period_tpe,), axis=-1)
            self.train_trend = np.concatenate((self.train_trend, self.train_trend_tpe,), axis=-1)

            self.test_closeness = np.concatenate((self.test_closeness, self.test_closeness_tpe,), axis=-1)
            self.test_period = np.concatenate((self.test_period, self.test_period_tpe,), axis=-1)
            self.test_trend = np.concatenate((self.test_trend, self.test_trend_tpe,), axis=-1)

        else:

            self.tpe_dim = None


    def st_map(self, zoom=11, style='mapbox://styles/rmetfc/ck1manozn0edb1dpmvtzle2cp', build_order=None):
        if self.dataset.node_station_info is None or len(self.dataset.node_station_info) == 0:
            raise ValueError('No station information found in dataset')

        import numpy as np
        import plotly
        from plotly.graph_objs import Scattermapbox, Layout

        mapboxAccessToken = "pk.eyJ1Ijoicm1ldGZjIiwiYSI6ImNrMW02YmwxbjAxN24zam9kNGVtMm5raWIifQ.FXKqZCxsFK-dGLLNdeRJHw"

        # os.environ['MAPBOX_API_KEY'] = mapboxAccessToken

        # lat_lng_name_list = [e[2:] for e in self.dataset.node_station_info]
        build_order = build_order or list(range(len(self.dataset.node_station_info)))

        color = ['rgb(255, 0, 0)' for _ in build_order]

        lat = np.array([float(e[2]) for e in self.dataset.node_station_info])[self.traffic_data_index]
        lng = np.array([float(e[3]) for e in self.dataset.node_station_info])[self.traffic_data_index]
        text = [str(e) for e in range(len(build_order))]
        if self.dataset.city:
            file_name = self.dataset.dataset + '-' + self.dataset.city + '.html'
        else:
            file_name = self.dataset.dataset+'.html'
        bikeStations = [Scattermapbox(
            lon=lng,
            lat=lat,
            text=text,
            mode='markers',
            marker=dict(
                size=6,
                #  color=['rgb(%s, %s, %s)' % (255,
                #                                               195 - e * 195 / max(build_order),
                #                                               195 - e * 195 / max(build_order)) for e in build_order],
                color=color,
                opacity=1,
            ))]

        layout = Layout(
            title='Bike Station Location & The latest built stations with deeper color',
            autosize=True,
            hovermode='closest',
            showlegend=False,
            mapbox=dict(
                accesstoken=mapboxAccessToken,
                bearing=0,
                center=dict(
                    lat=np.median(lat),
                    lon=np.median(lng)
                ),
                pitch=0,
                zoom=zoom,
                style=style
            ),
        )

        fig = dict(data=bikeStations, layout=layout)
        plotly.offline.plot(fig, filename=file_name)

    def make_concat(self, node='all', is_train=True):
        """A function to concatenate all closeness, period and trend history data to use as inputs of models.

        Args:
            node (int or ``'all'``): To specify the index of certain node. If set to ``'all'``, return the concatenation
                result of all nodes. If set to an integer, it will be the index of the selected node. Default: ``'all'``
            is_train (bool): If set to ``True``, ``train_closeness``, ``train_period``, and ``train_trend`` will be
                concatenated. If set to ``False``, ``test_closeness``, ``test_period``, and ``test_trend`` will be
                concatenated. Default: True

        Returns:
            np.ndarray: Function returns an ndarray with shape as
            [time_slot_num, ``station_number``, ``closeness_len`` + ``period_len`` + ``trend_len``, 1],
            and time_slot_num is the temporal length of train set data if ``is_train`` is ``True``
            or the temporal length of test set data if ``is_train`` is ``False``.
            On the second dimension, data are arranged as
            ``earlier closeness -> later closeness -> earlier period -> later period -> earlier trend -> later trend``.
        """

        if is_train:
            length = len(self.train_y)
            closeness = self.train_closeness
            period = self.train_period
            trend = self.train_trend
        else:
            length = len(self.test_y)
            closeness = self.test_closeness
            period = self.test_period
            trend = self.test_trend
        if node == 'all':
            node = list(range(self.station_number))
        else:
            node = [node]
        history = np.zeros([length, len(node), self.closeness_len + self.period_len + self.trend_len])
        for i in range(len(node)):
            for c in range(self.closeness_len):
                history[:, i, c] = closeness[:, node[i], c, -1]
            for p in range(self.period_len):
                history[:, i, self.closeness_len + p] = period[:, node[i], p, -1]
            for t in range(self.trend_len):
                history[:, i, self.closeness_len + self.period_len + t] = trend[:, node[i], t, -1]
        history = np.expand_dims(history, 3)
        return history


class TransferDataLoader(object):

    def __init__(self, sd_params, td_params, model_params, td_data_length=None):

        if td_data_length:
            td_params.update({'train_data_length': td_data_length})

        self.sd_loader = NodeTrafficLoader(**sd_params, **model_params)
        self.td_loader = NodeTrafficLoader(**td_params, **model_params)

        td_params.update({'train_data_length': '180'})
        self.fake_td_loader = NodeTrafficLoader(**td_params, **model_params)

    def traffic_sim(self):

        assert self.sd_loader.daily_slots == self.td_loader.daily_slots

        similar_record = []

        for i in range(0, self.sd_loader.train_data.shape[0] - self.td_loader.train_data.shape[0],
                       int(self.sd_loader.daily_slots)):

            sim = cosine_similarity(self.td_loader.train_data.transpose(),
                                    self.sd_loader.train_data[i:i + self.td_loader.train_data.shape[0]].transpose())

            max_sim, max_index = np.max(sim, axis=1), np.argmax(sim, axis=1)

            if len(similar_record) == 0:
                similar_record = [[max_sim[e], max_index[e], i, i + self.td_loader.train_data.shape[0]]
                                  for e in range(len(max_sim))]
            else:
                for index in range(len(similar_record)):
                    if similar_record[index][0] < max_sim[index]:
                        similar_record[index] = [max_sim[index], max_index[index], i,
                                                 i + self.td_loader.train_data.shape[0]]

        return similar_record

    def traffic_sim_fake(self):

        assert self.sd_loader.daily_slots == self.fake_td_loader.daily_slots

        similar_record = []

        for i in range(0, self.sd_loader.train_data.shape[0] - self.fake_td_loader.train_data.shape[0],
                       int(self.sd_loader.daily_slots)):

            sim = cosine_similarity(self.fake_td_loader.train_data.transpose(),
                                    self.sd_loader.train_data[
                                    i:i + self.fake_td_loader.train_data.shape[0]].transpose())

            max_sim, max_index = np.max(sim, axis=1), np.argmax(sim, axis=1)

            if len(similar_record) == 0:
                similar_record = [[max_sim[e], max_index[e], i, i + self.fake_td_loader.train_data.shape[0]]
                                  for e in range(len(max_sim))]
            else:
                for index in range(len(similar_record)):
                    if similar_record[index][0] < max_sim[index]:
                        similar_record[index] = [max_sim[index], max_index[index], i,
                                                 i + self.td_loader.train_data.shape[0]]

        return similar_record

    def checkin_sim(self):

        from sklearn.metrics.pairwise import cosine_similarity

        td_checkin = np.array([e[0] for e in self.td_loader.dataset.data['ExternalFeature']['CheckInFeature']]
                              )[self.td_loader.traffic_data_index]
        sd_checkin = np.array([e[0] for e in self.sd_loader.dataset.data['ExternalFeature']['CheckInFeature']]
                              )[self.sd_loader.traffic_data_index]

        td_checkin = td_checkin / (np.max(td_checkin, axis=1, keepdims=True) + 0.0001)
        sd_checkin = sd_checkin / (np.max(sd_checkin, axis=1, keepdims=True) + 0.0001)

        # cs = cosine_similarity(td_checkin, sd_checkin)

        # similar_record = [[e[np.argmax(e)], np.argmax(e), ] for e in cs]

        similar_record = []
        for td_index in range(len(td_checkin)):
            tmp_sim_record = []
            for sd_index in range(len(sd_checkin)):
                r, p = pearsonr(td_checkin[td_index], sd_checkin[sd_index])
                tmp_sim_record.append([r, sd_index,
                                       len(self.sd_loader.train_y) - len(self.td_loader.train_y),
                                       len(self.sd_loader.train_y)])
            similar_record.append(max(tmp_sim_record, key=lambda x: x[0]))

        return similar_record

    def checkin_sim_sd(self):

        sd_checkin = np.array([e[0] for e in self.sd_loader.dataset.data['ExternalFeature']['CheckInFeature']]
                              )[self.sd_loader.traffic_data_index]
        sd_checkin = sd_checkin / (np.max(sd_checkin, axis=1, keepdims=True) + 0.0001)

        cs = cosine_similarity(sd_checkin, sd_checkin) - np.eye(sd_checkin.shape[0])

        return np.array([np.argmax(e) for e in cs], np.int32)

    def poi_sim(self):

        from sklearn.metrics.pairwise import cosine_similarity

        td_checkin = np.array([e[1] for e in self.td_loader.dataset.data['ExternalFeature']['CheckInFeature']]
                              )[self.td_loader.traffic_data_index]
        sd_checkin = np.array([e[1] for e in self.sd_loader.dataset.data['ExternalFeature']['CheckInFeature']]
                              )[self.sd_loader.traffic_data_index]

        return [[e[np.argmax(e)], np.argmax(e), ] for e in cosine_similarity(td_checkin, sd_checkin)]




from .EarlyStopping import EarlyStopping
from .EarlyStopping import EarlyStoppingTTest

from .MiniBatchTrain import MiniBatchTrain, MiniBatchTrainMultiData, MiniBatchFeedDict

from .LossFunction import *


import xgboost as xgb
import numpy as np


class XGBoost():
    """
    XGBoost is an optimized distributed gradient boosting machine learning algorithm.

    Args:
        *n_estimators (int): Number of boosting iterations. Default: 10
        *max_depth (int): Maximum tree depth for base learners. Default: 5
        *verbosity (int): The degree of verbosity. Valid values are 0 (silent) - 3 (debug). Default: 0
        *objective (string or callable):
            Specify the learning task and the corresponding learning objective or
            a custom objective function to be used. Default: ``'reg:squarederror'``
        *eval_metric (str, list of str, or callable, optional):
            If a str, should be a built-in evaluation metric to use. See more in
            `API Reference of XGBoost Library <https://xgboost.readthedocs.io/en/latest/python/python_api.html>`_.
            Default: ``'rmse'``
    """
    def __init__(self, n_estimators=10, max_depth=5, verbosity=0, objective='reg:squarederror', eval_metric='rmse'):
        
        self.param = {
            'max_depth': max_depth,
            'verbosity': verbosity,
            'objective': objective,
            'eval_metric': eval_metric
        }
        self.n_estimators = n_estimators

    def fit(self, X, y):
        '''
        Training method. 

        Args:
            X(np.ndarray/scipy.sparse/pd.DataFrame/dt.Frame): The training input samples.
            y(np.ndarray, optional): The target values of training samples.
        '''
        train_matrix = xgb.DMatrix(X, label=y)
        self.model = xgb.train(self.param, train_matrix, self.n_estimators)

    def predict(self, X):
        '''
        Prediction method.
        
        :Returns: Predicted values with shape as [time_slot_num, node_num, 1].
        :Return type: np.ndarray
        '''
        test_matrix = xgb.DMatrix(X)
        return self.model.predict(test_matrix)





from .HM import HM
from .ARIMA import ARIMA

try:
    from .HMM import HMM
except ModuleNotFoundError:
    print('HMM not installed')

from .XGBoost import XGBoost
# from .GBRT import GBRT

from .DeepST import DeepST
from .ST_ResNet import ST_ResNet

from .STMeta import STMeta

from .DCRNN import DCRNN

from .ST_MGCN import ST_MGCN
from .GeoMAN import GeoMAN
from .AGCRN import AGCRN
from .ASTGCN import make_model
from .GMAN import GMAN
from .GraphWaveNet import gwnet
from .STGCN import build_model
from .STSGCN import stsgcn



import numpy as np
import pandas as pd
import statsmodels.api as sm

import warnings
warnings.filterwarnings("ignore")


class ARIMA(object):
    """ARIMA is a generalization of an ARMA (Autoregressive Moving Average) model, used in predicting 
    future points in time series analysis.

    Since there may be three kinds of series data as closeness, period and trend history, this class
    trains three different ARIMA models for each node according  to the three kinds of history data, 
    and returns average of the predicted values by the models in prediction.

    Args:
        time_sequence(array_like): The observation value of time_series.
        order(iterable): It stores the (p, d, q) orders of the model for the number of AR parameters
            , differences, MA parameters. If set to None, ARIMA class will calculate the orders for 
            each series based on max_ar, max_ma and max_d. Default: None
        seasonal_order(iterable): It stores the (P,D,Q,s) order of the seasonal ARIMA model for the
            AR parameters, differences, MA parameters, and periodicity. `s` is an integer giving the 
            periodicity (number of periods in season).
        max_ar(int): Maximum number of AR lags to use. Default: 6
        max_ma(int): Maximum number of MA lags to use. Default: 4
        max_d(int): Maximum number of degrees of differencing. Default: 2

    Attribute:
        order(iterable): (p, d, q) orders for ARIMA model. 
        seasonal_order(iterable): (P,D,Q,s) order for seasonal ARIMA model. 
        model_res(): Fit method for likelihood based models. 
    """

    def __init__(self, time_sequence, order=None, seasonal_order=(0, 0, 0, 0), max_ar=6, max_ma=4, max_d=2):

        self.seasonal_order = seasonal_order
        auto_order = self.get_order(time_sequence, order, max_ar=max_ar, max_ma=max_ma, max_d=max_d)
        model = sm.tsa.SARIMAX(time_sequence, order=auto_order, seasonal_order=self.seasonal_order)
        model_res = model.fit(disp=False)
        self.order = auto_order
        self.model_res = model_res

    def get_order(self, series, order=None, max_ar=6, max_ma=2, max_d=2):
        '''
        If order is None, it simply returns order, otherwise, it calculates the (p, d, q) orders 
        for the series data based on max_ar, max_ma and max_d.
        '''

        def stationary(series):
            t = ARIMA.adf_test(series, verbose=False)
            if t[0] < t[4]['1%']:
                return True
            else:
                return False

        if order is None:
            order_i = 0
            while not stationary(np.diff(series, order_i)):
                order_i += 1
                if order_i > max_d:
                    break
            order = sm.tsa.stattools.arma_order_select_ic(np.diff(series, order_i),
                                                          max_ar=max_ar, max_ma=max_ma,
                                                          ic=['aic']).aic_min_order
            order = list(order)
            order.insert(1, order_i)

        return order

    @staticmethod
    def adf_test(time_series, max_lags=None, verbose=True):
        '''
        Augmented Dickey–Fuller test. The Augmented Dickey-Fuller test can be used to test for 
        a unit root in a univariate process in the presence of serial correlation.
        '''
        t = sm.tsa.stattools.adfuller(time_series, maxlag=max_lags)
        if verbose:
            output = pd.DataFrame(
                index=['Test Statistic Value',
                       "p-value",
                       "Lags Used",
                       "Number of Observations Used",
                       "Critical Value(1%)",
                       "Critical Value(5%)", "Critical Value(10%)"], columns=['value'])
            output['value']['Test Statistic Value'] = t[0]
            output['value']['p-value'] = t[1]
            output['value']['Lags Used'] = t[2]
            output['value']['Number of Observations Used'] = t[3]
            output['value']['Critical Value(1%)'] = t[4]['1%']
            output['value']['Critical Value(5%)'] = t[4]['5%']
            output['value']['Critical Value(10%)'] = t[4]['10%']
            print(output)
        return t

    def predict(self, time_sequences, forecast_step=1):
        '''
        Argues:
            time_sequences: The input time_series features.
            forecast_step: The number of predicted future steps. Default: 1
        
        :return: Prediction results with shape of (len(time_sequence)/forecast_step,forecast_step=,1).
        :type: np.ndarray
        '''
        result = []
        """ origin predict method, output shape is [math.ceil(len(time_sequences) / forecast_step), forecast_step]
        for i in range(0, len(time_sequences), forecast_step):
            fs = forecast_step if ((i + forecast_step) < len(time_sequences)) else (len(time_sequences) - i)
            model = sm.tsa.SARIMAX(time_sequences[i], order=self.order, seasonal_order=self.seasonal_order)
            model_res = model.filter(self.model_res.params)
            p = model_res.forecast(fs).reshape([-1, 1])
            result.append(p)
        """
        # new predict method, output shape is [len(time_sequences), forecast_step]
        for i in range(len(time_sequences)):
            model = sm.tsa.SARIMAX(time_sequences[i], order=self.order, seasonal_order=self.seasonal_order)
            model_res = model.filter(self.model_res.params)
            p = model_res.forecast(forecast_step)
            p = p.reshape([-1, forecast_step])
            result.append(p)
        if forecast_step != 1:
            result = np.concatenate(result, axis=0)
        return np.array(result, dtype=np.float32)


import numpy as np
import torch
import tensorflow as tf
import mxnet as mx

def masked_mse(preds, labels, null_val=np.nan):
    if np.isnan(null_val):
        mask = ~torch.isnan(labels)
    else:
        mask = (labels != null_val)
    mask = mask.float()
    # print(mask.sum())
    # print(mask.shape[0]*mask.shape[1]*mask.shape[2])
    mask /= torch.mean((mask))
    mask = torch.where(torch.isnan(mask), torch.zeros_like(mask), mask)
    loss = (preds - labels)**2
    loss = loss * mask
    loss = torch.where(torch.isnan(loss), torch.zeros_like(loss), loss)
    return torch.mean(loss)


def masked_rmse(preds, labels, null_val=np.nan):
    return torch.sqrt(masked_mse(preds=preds, labels=labels,
                                 null_val=null_val))


def masked_mae(preds, labels, null_val=np.nan):
    if np.isnan(null_val):
        mask = ~torch.isnan(labels)
    else:
        mask = (labels != null_val)
    mask = mask.float()
    mask /= torch.mean((mask))
    mask = torch.where(torch.isnan(mask), torch.zeros_like(mask), mask)
    loss = torch.abs(preds - labels)
    loss = loss * mask
    loss = torch.where(torch.isnan(loss), torch.zeros_like(loss), loss)
    return torch.mean(loss)


def masked_mape(preds, labels, null_val=np.nan):
    if np.isnan(null_val):
        mask = ~torch.isnan(labels)
    else:
        mask = (labels != null_val)
    mask = mask.float()
    mask /= torch.mean((mask))
    mask = torch.where(torch.isnan(mask), torch.zeros_like(mask), mask)
    loss = torch.abs(preds-labels)/labels
    loss = loss * mask
    loss = torch.where(torch.isnan(loss), torch.zeros_like(loss), loss)
    return torch.mean(loss)


def mae_loss(pred, label):
    mask = tf.not_equal(label, 0)
    mask = tf.cast(mask, tf.float32)
    mask /= tf.reduce_mean(mask)
    # mask = tf.compat.v2.where(
    #     condition = tf.math.is_nan(mask), x = 0., y = mask)
    mask = tf.compat.v1.where(
        condition=tf.math.is_nan(mask), x=tf.zeros_like(mask), y=mask)
    loss = tf.abs(tf.subtract(pred, label))
    loss *= mask
    # loss = tf.compat.v2.where(
    #     condition = tf.math.is_nan(loss), x = 0., y = loss)
    loss = tf.compat.v1.where(
        condition=tf.math.is_nan(loss), x=tf.zeros_like(mask), y=loss)
    loss = tf.reduce_mean(loss)
    return loss

def mask_np(array, null_val):
    '''
    function for STSGCN
    '''
    if np.isnan(null_val):
        return (~np.isnan(null_val)).astype('float32')
    else:
        return np.not_equal(array, null_val).astype('float32')


def masked_mse_np(y_true, y_pred, null_val=np.nan):
    '''
    function for STSGCN
    '''
    mask = mask_np(y_true, null_val)
    mask /= mask.mean()
    mse = (y_true - y_pred) ** 2
    return np.mean(np.nan_to_num(mask * mse))


def masked_mae_np(y_true, y_pred, null_val=np.nan):
    '''
    function for STSGCN
    '''
    mask = mask_np(y_true, null_val)
    mask /= mask.mean()
    mae = np.abs(y_true - y_pred)
    return np.mean(np.nan_to_num(mask * mae))


def masked_mape_np(y_true, y_pred, null_val=np.nan):
    '''
    function for STSGCN
    '''
    with np.errstate(divide='ignore', invalid='ignore'):
        mask = mask_np(y_true, null_val)
        mask /= mask.mean()
        mape = np.abs((y_pred - y_true) / y_true)
        mape = np.nan_to_num(mask * mape)
        return np.mean(mape) * 100


def masked_mae_loss(mask_value):
    '''
    function for AGCRN
    '''
    def loss(preds, labels):
        mae = MAE_torch(pred=preds, true=labels, mask_value=mask_value)
        return mae
    return loss

def MAE_torch(pred, true, mask_value=None):
    if mask_value != None:
        mask = torch.gt(true, mask_value)
        pred = torch.masked_select(pred, mask)
        true = torch.masked_select(true, mask)
    return torch.mean(torch.abs(true-pred))



def huber_loss(data, label, rho=1):
    '''
    Parameters
    ----------
    data: mx.sym.var, shape is (B, T', N)

    label: mx.sym.var, shape is (B, T', N)

    rho: float

    Returns
    ----------
    loss: mx.sym
    '''

    loss = mx.sym.abs(data - label)
    loss = mx.sym.where(loss > rho, loss - 0.5 * rho,
                        (0.5 / rho) * mx.sym.square(loss))
    loss = mx.sym.MakeLoss(loss)
    return loss



from dateutil.parser import parse
from chinese_calendar import is_workday
from workalendar.usa import NewYork, DistrictOfColumbia, Illinois

america_public_holiday = ['01-01', '01-02', '01-16', '02-12', '02-13', '02-20', '05-29', '07-04', '09-04',
                          '10-09', '11-10', '11-11', '11-23', '12-25']


def is_work_day_america(date, city):
    """
    Args:
        date(string or datetime): e.g. 2019-01-01

    Return:
        True if date is not holiday in America,
        otherwise return False.
    """
    if type(date) is str:
        date = parse(date)

    if city == "Chicago":
        workday = Illinois()
    elif city == "NYC":
        workday = NewYork()
    elif city == "DC":
        workday = DistrictOfColumbia()
    else:
        raise ValueError("can't parse holiday in {}.".format(city))
    return workday.is_working_day(date)


def is_work_day_china(date, city):
    """
    Args:
        date(string or datetime): e.g. 2019-01-01

    Return:
        True if date is not holiday in China,
        otherwise return False.
    """
    if type(date) is str:
        date = parse(date)
    return is_workday(date)


def is_valid_date(date_str):
    """
    Args:
        date_str(string): e.g. 2019-01-01

    Return:
        True if date_str is valid date,
        otherwise return False.
    """
    try:
        date = parse(date_str)
    except:
        return False

    year = date.year
    month = date.month
    day = date.day

    isRunNian = False
    if year % 4 == 0 and year % 100 != 0 and year % 400 == 0:
        isRunNian = True

    if month < 1 or month > 12:
        return False

    pingnian_month = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    runnian_month = [0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

    if isRunNian:
        if day < 1 or day > pingnian_month[month]:
            return False
    else:
        if day < 1 or day > runnian_month[month]:
            return False

    return True



import tensorflow as tf

from tensorflow.python.framework import dtypes
from tensorflow.python.keras import backend as K
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.ops import array_ops, linalg_ops, math_ops


def _generate_dropout_mask(ones, rate, training=None, count=1):
    def dropped_inputs():
        return K.dropout(ones, rate)

    if count > 1:
        return [
            K.in_train_phase(dropped_inputs, ones, training=training)
            for _ in range(count)
        ]
    return K.in_train_phase(dropped_inputs, ones, training=training)


class GCLSTMCell(tf.keras.layers.LSTMCell):

    """
    GCLSTMCell is one of our implemented ST-RNN models in handling the spatial and temporal features.
    We performed GCN on both LSTM inputs and hidden-states. The code is inherited from tf.keras.layers.LSTMCell,
    thus it can be used almost the same as LSTMCell except that you need to provide the GCN parameters
    in the __init__ function.

    Args:
        units(int): number of units of LSTM
        num_node(int): number of nodes in the graph
        laplacian_matrix(ndarray): laplacian matrix used in GCN, with shape [num_node, num_node]
        gcn_k(int): highest order of Chebyshev Polynomial approximation in GCN
        gcn_l(int): number of GCN layers
        kwargs: other parameters supported by LSTMCell, such as activation, kernel_initializer ... and so on.
    """

    def __init__(self, units, num_node, laplacian_matrix, gcn_k=1, gcn_l=1, **kwargs):

        super().__init__(units, **kwargs)

        self._units = units
        self._num_node = num_node
        self._gcn_k = gcn_k
        self._gcn_l = gcn_l
        self._laplacian_matrix = laplacian_matrix

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        super(GCLSTMCell, self).build(input_shape)
        input_dim = input_shape[-1]
        self.kernel = self.add_weight(
            shape=(input_dim * (self._gcn_k + 1), self.units * 4),
            name='kernel',
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint)
        self.recurrent_kernel = self.add_weight(
            shape=(self.units * (self._gcn_k + 1), self.units * 4),
            name='recurrent_kernel',
            initializer=self.recurrent_initializer,
            regularizer=self.recurrent_regularizer,
            constraint=self.recurrent_constraint)

    def kth_cheby_ploy(self, k, tk1=None, tk2=None):
        if k == 0:
            return linalg_ops.eye(self._num_node, dtype=dtypes.float32)
        elif k == 1:
            return self._laplacian_matrix
        elif k > 1:
            return math_ops.matmul(2 * self._laplacian_matrix, tk1) - tk2

    def call(self, inputs, states, training=None):

        if 0 < self.dropout < 1 and self._dropout_mask is None:
            self._dropout_mask = _generate_dropout_mask(
                array_ops.ones_like(inputs),
                self.dropout,
                training=training,
                count=4)
        if (0 < self.recurrent_dropout < 1 and
                self._recurrent_dropout_mask is None):
            self._recurrent_dropout_mask = _generate_dropout_mask(
                array_ops.ones_like(states[0]),
                self.recurrent_dropout,
                training=training,
                count=4)

        input_dim = inputs.get_shape()[-1].value

        # dropout matrices for input units
        dp_mask = self._dropout_mask
        # dropout matrices for recurrent units
        rec_dp_mask = self._recurrent_dropout_mask

        h_tm1 = states[0]  # previous memory state
        c_tm1 = states[1]  # previous carry state

        if 0. < self.dropout < 1.:
            inputs *= dp_mask[0]
        if 0. < self.recurrent_dropout < 1.:
            h_tm1 *= rec_dp_mask[0]

        # inputs has shape: [batch * num_node, input_dim]
        # h_tm1 has shape: [batch * num_node, units]
        inputs_before_gcn = tf.reshape(tf.transpose(tf.reshape(inputs, [-1, self._num_node, input_dim]),
                                                    [1, 0, 2]), [self._num_node, -1])
        h_tm1_before_gcn = tf.reshape(tf.transpose(tf.reshape(h_tm1, [-1, self._num_node, self._units]),
                                                   [1, 0, 2]), [self._num_node, -1])

        t = []
        inputs_after_gcn = list()
        h_tm1_after_gcn = list()
        for i in range(0, self._gcn_k + 1):
            t.append(self.kth_cheby_ploy(k=i, tk1=None if i < 1 else t[i - 1], tk2=None if i < 2 else t[i - 2]))
            inputs_after_gcn.append(tf.matmul(t[-1], inputs_before_gcn))
            h_tm1_after_gcn.append(tf.matmul(t[-1], h_tm1_before_gcn))

        inputs_after_gcn = tf.reshape(inputs_after_gcn, [self._gcn_k + 1, self._num_node, -1, input_dim])
        h_tm1_after_gcn = tf.reshape(h_tm1_after_gcn, [self._gcn_k + 1, self._num_node, -1, self._units])

        inputs_after_gcn = tf.reshape(tf.transpose(inputs_after_gcn, [2, 1, 0, 3]), [-1, (self._gcn_k + 1) * input_dim])
        h_tm1_after_gcn = tf.reshape(tf.transpose(h_tm1_after_gcn, [2, 1, 0, 3]), [-1, (self._gcn_k + 1) * self.units])

        z = K.dot(inputs_after_gcn, self.kernel)
        z += K.dot(h_tm1_after_gcn, self.recurrent_kernel)
        if self.use_bias:
            z = K.bias_add(z, self.bias)

        z0 = z[:, :self.units]
        z1 = z[:, self.units:2 * self.units]
        z2 = z[:, 2 * self.units:3 * self.units]
        z3 = z[:, 3 * self.units:]

        z = (z0, z1, z2, z3)
        c, o = self._compute_carry_and_output_fused(z, c_tm1)

        h = o * self.activation(c)
        return h, [h, c]



import numpy as np
import tensorflow as tf
import heapq

# from UCTB.preprocess import Normalizer, SplitData

from math import radians, cos, sin, asin, sqrt
from scipy.stats import pearsonr
from scipy.sparse.linalg import eigs

class GraphGenerator():
    '''
    This class is used to build graphs. 
    Adajacent matrix and lapalace matrix will be stored in self.AM and self.LM.

    Args:
        data_loader(NodeTrafficLoader): data_loader object.
        graph (str): Types of graphs used in neural methods. Graphs should be a subset of { ``'Correlation'``,
            ``'Distance'``, ``'Interaction'``, ``'Line'``, ``'Neighbor'``, ``'Transfer'`` } and concatenated by ``'-'``,
            and *dataset* should have data of selected graphs. Default: ``'Correlation'``
        threshold_distance (float): Used in building of distance graph. If distance of two nodes in meters is larger
            than ``threshold_distance``, the corresponding position of the distance graph will be 1 and otherwise
            0.the corresponding Default: 1000
        threshold_correlation (float): Used in building of correlation graph. If the Pearson correlation coefficient is
            larger than ``threshold_correlation``, the corresponding position of the correlation graph will be 1
            and otherwise 0. Default: 0
        threshold_interaction (float): Used in building of interatction graph. If in the latest 12 months, the number of
            times of interaction between two nodes is larger than ``threshold_interaction``, the corresponding position
            of the interaction graph will be 1 and otherwise 0. Default: 500

    Attributes:
        AM (array): Adajacent matrices of graphs.
        LM (array): Laplacian matrices of graphs.
    '''

    def __init__(self,
                 data_loader,
                 graph="Correlation",
                 threshold_distance=1000,
                 threshold_correlation=0,
                 threshold_interaction=500, **kwargs):
        self.AM = []
        self.LM = []
        self.threshold_distance = threshold_distance
        self.threshold_correlation = threshold_correlation
        self.threshold_interaction = threshold_interaction

        self.dataset = data_loader.dataset
        self.train_data = data_loader.train_data
        self.traffic_data_index = data_loader.traffic_data_index
        self.train_test_ratio = data_loader.train_test_ratio
        self.daily_slots = 24 * 60 / self.dataset.time_fitness

        # build_graph
        for graph_name in graph.split('-'):
            AM, LM = self.build_graph(graph_name)
            if AM is not None:
                self.AM.append(AM)
            if LM is not None:
                self.LM.append(LM)
        
        self.AM = np.array(self.AM, dtype=np.float32)
        self.LM = np.array(self.LM, dtype=np.float32)
        # print (self.LM.shape[:])

    def build_graph(self, graph_name):
        AM, LM = None, None
        if graph_name.lower() == 'distance':
            lat_lng_list = np.array([[float(e1) for e1 in e[2:4]]
                                     for e in self.dataset.node_station_info])
            AM = self.distance_adjacent(lat_lng_list[self.traffic_data_index],
                                                threshold=float(self.threshold_distance))
            LM = self.adjacent_to_laplacian(AM)

        if graph_name.lower() == 'interaction':
            monthly_interaction = self.dataset.node_monthly_interaction[:, self.traffic_data_index, :][:, :,
                                                                                                       self.traffic_data_index]

            monthly_interaction, _ = SplitData.split_data(
                monthly_interaction, self.train_test_ratio)

            annually_interaction = np.sum(monthly_interaction[-12:], axis=0)
            annually_interaction = annually_interaction + annually_interaction.transpose()

            AM = self.interaction_adjacent(annually_interaction,
                                                   threshold=float(self.threshold_interaction))
            LM = self.adjacent_to_laplacian(AM)

        if graph_name.lower() == 'correlation':
            AM = self.correlation_adjacent(self.train_data[-30 * int(self.daily_slots):],
                                                   threshold=float(self.threshold_correlation))
            LM = self.adjacent_to_laplacian(AM)

        if graph_name.lower() == 'neighbor':
            LM = self.adjacent_to_laplacian(
                self.dataset.data.get('contribute_data').get('graph_neighbors'))

        if graph_name.lower() == 'line':
            LM = self.adjacent_to_laplacian(
                self.dataset.data.get('contribute_data').get('graph_lines'))
            LM = LM[self.traffic_data_index]
            LM = LM[:, self.traffic_data_index]

        if graph_name.lower() == 'transfer':
            LM = self.adjacent_to_laplacian(
                self.dataset.data.get('contribute_data').get('graph_transfer'))
        return AM, LM

    @staticmethod
    def haversine(lat1, lon1, lat2, lon2):
        """
        Calculate the great circle distance between two points
        on the earth (specified in decimal degrees)
        """
        lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

        # haversine
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
        c = 2 * asin(sqrt(a))
        r = 6371

        return c * r * 1000

    @staticmethod
    def correlation_adjacent(traffic_data, threshold):
        '''
        Calculate correlation graph based on pearson coefficient.

        Args:
            traffic_data(ndarray): numpy array with shape [sequence_length, num_node].
            threshold(float): float between [-1, 1], nodes with Pearson Correlation coefficient
                larger than this threshold will be linked together.
        '''
        adjacent_matrix = np.zeros([traffic_data.shape[1], traffic_data.shape[1]])
        for i in range(traffic_data.shape[1]):
            for j in range(traffic_data.shape[1]):
                r, p_value = pearsonr(traffic_data[:, i], traffic_data[:, j])
                adjacent_matrix[i, j] = 0 if np.isnan(r) else r
        adjacent_matrix = (adjacent_matrix >= threshold).astype(np.float32)
        return adjacent_matrix

    def distance_adjacent(self, lat_lng_list, threshold):
        '''
        Calculate distance graph based on geographic distance.

        Args:
            lat_lng_list(list): A list of geographic locations. The format of each element
                    in the list is [latitude, longitude].
            threshold(float): (meters) nodes with geographic distacne smaller than this 
                threshold will be linked together.
        '''
        adjacent_matrix = np.zeros([len(lat_lng_list), len(lat_lng_list)])
        for i in range(len(lat_lng_list)):
            for j in range(len(lat_lng_list)):
                adjacent_matrix[i][j] = self.haversine(lat_lng_list[i][0], lat_lng_list[i][1],
                                                                lat_lng_list[j][0], lat_lng_list[j][1])
        adjacent_matrix = (adjacent_matrix <= threshold).astype(np.float32)
        return adjacent_matrix

    @staticmethod
    def interaction_adjacent(interaction_matrix, threshold):
        '''
        Binarize interaction_matrix based on threshold.

        Args:

            interaction_matrix(ndarray): with shape [num_node, num_node], where each 
                element represents the number of interactions during a certain time,
                    e.g. 6 monthes, between the corresponding nodes.
            threshold(float or int): nodes with number of interactions between them
                    greater than this threshold will be linked together.
        '''
        return (interaction_matrix >= threshold).astype(np.float32)      

    @staticmethod
    def adjacent_to_laplacian(adjacent_matrix):
        '''
        Turn adjacent_matrix into Laplace matrix.
        '''
        adjacent_matrix -= np.diag(np.diag(adjacent_matrix))
        diagonal_matrix = np.diag(np.sum(adjacent_matrix, axis=0) ** -0.5)
        diagonal_matrix[np.isinf(diagonal_matrix)] = 0
        laplacian_matrix = np.eye(len(adjacent_matrix)) - np.dot(np.dot(diagonal_matrix, adjacent_matrix),
                                                                    diagonal_matrix)
        laplacian_matrix = 2 * laplacian_matrix / np.max(laplacian_matrix) - np.eye(len(adjacent_matrix))
        return laplacian_matrix



def scaled_Laplacian_ASTGCN(W):
    '''
    compute \tilde{L}

    Parameters
    ----------
    W(np.ndarray): shape is (num_node, num_node).

    Returns
    ----------
    scaled_Laplacian_ASTGCN: np.ndarray, shape (num_node, num_node)

    '''

    assert W.shape[0] == W.shape[1]

    D = np.diag(np.sum(W, axis=1))

    L = D - W

    lambda_max = eigs(L, k=1, which='LR')[0].real

    return (2 * L) / lambda_max - np.identity(W.shape[0])


def scaled_laplacian_STGCN(W):
    '''
    Normalized graph Laplacian function.

    Args:
        W(np.ndarray): [num_node, num_node], weighted adjacency matrix of G.
    :return: Scaled laplacian matrix.
    :type: np.matrix, [num_node, num_node].
    '''
    # d ->  diagonal degree matrix
    n, d = np.shape(W)[0], np.sum(W, axis=1)
    # L -> graph Laplacian
    L = -W
    L[np.diag_indices_from(L)] = d
    for i in range(n):
        for j in range(n):
            if (d[i] > 0) and (d[j] > 0):
                L[i, j] = L[i, j] / np.sqrt(d[i] * d[j])
    # lambda_max \approx 2.0, the largest eigenvalues of L.
    lambda_max = eigs(L, k=1, which='LR')[0][0].real
    return np.mat(2 * L / lambda_max - np.identity(n))



from os.path import join as pjoin
import tensorflow as tf
import numpy as np
import time
from UCTB.preprocess import SplitData
from UCTB.model.STGCN import build_model, gen_batch


def model_save(sess, global_steps, model_name, save_path='./output/models/'):
    '''
    Save the checkpoint of trained model.
    :param sess: tf.Session().
    :param global_steps: tensor, record the global step of training in epochs.
    :param model_name: str, the name of saved model.
    :param save_path: str, the path of saved model.
    :return:
    '''
    saver = tf.train.Saver(max_to_keep=3)
    prefix_path = saver.save(sess, pjoin(
        save_path, model_name), global_step=global_steps)
    print(f'<< Saving model to {prefix_path} ...')


def model_train(inputs, blocks, args, sum_path='./output/models'):
    '''
    Train the base model.
    :param inputs: instance of class Dataset, data source for training.
    :param blocks: list, channel configs of st_conv blocks.
    :param args: instance of class argparse, args for training.
    '''
    n, n_his, n_pred = args.n_route, args.n_his, args.n_pred
    print(n_his)
    Ks, Kt = args.ks, args.kt
    batch_size, epoch, inf_mode, opt = args.batch_size, args.epoch, args.inf_mode, args.opt

    # Placeholder for model training
    x = tf.placeholder(tf.float32, [None, n_his + 1, n, 1], name='data_input')
    keep_prob = tf.placeholder(tf.float32, name='keep_prob')

    # Define model loss
    train_loss, pred = build_model(x, n_his, Ks, Kt, blocks, keep_prob)
    tf.summary.scalar('train_loss', train_loss)
    copy_loss = tf.add_n(tf.get_collection('copy_loss'))
    tf.summary.scalar('copy_loss', copy_loss)

    # Learning rate settings
    global_steps = tf.Variable(0, trainable=False)
    len_train = inputs.get_len('train')
    if len_train % batch_size == 0:
        epoch_step = len_train / batch_size
    else:
        epoch_step = int(len_train / batch_size) + 1
    # Learning rate decay with rate 0.7 every 5 epochs.
    lr = tf.train.exponential_decay(
        args.lr, global_steps, decay_steps=5 * epoch_step, decay_rate=0.7, staircase=True)
    tf.summary.scalar('learning_rate', lr)
    step_op = tf.assign_add(global_steps, 1)
    with tf.control_dependencies([step_op]):
        if opt == 'RMSProp':
            train_op = tf.train.RMSPropOptimizer(lr).minimize(train_loss)
        elif opt == 'ADAM':
            train_op = tf.train.AdamOptimizer(lr).minimize(train_loss)
        else:
            raise ValueError(f'ERROR: optimizer "{opt}" is not defined.')

    merged = tf.summary.merge_all()

    with tf.Session() as sess:
        writer = tf.summary.FileWriter(pjoin(sum_path, 'train'), sess.graph)
        sess.run(tf.global_variables_initializer())

        if inf_mode == 'sep':
            # for inference mode 'sep', the type of step index is int.
            step_idx = n_pred - 1
            tmp_idx = [step_idx]
            min_val = min_va_val = np.array([4e1, 1e5, 1e5])
        elif inf_mode == 'merge':
            # for inference mode 'merge', the type of step index is np.ndarray.
            step_idx = tmp_idx = np.arange(3, n_pred + 1, 3) - 1
            min_val = min_va_val = np.array([4e1, 1e5, 1e5] * len(step_idx))
        else:
            raise ValueError(f'ERROR: test mode "{inf_mode}" is not defined.')

        for i in range(epoch):
            start_time = time.time()
            for j, x_batch in enumerate(
                    gen_batch(inputs.get_data('train'), batch_size, dynamic_batch=True, shuffle=True)):
                summary, _ = sess.run([merged, train_op], feed_dict={
                                      x: x_batch[:, 0:n_his + 1, :, :], keep_prob: 1.0})
                writer.add_summary(summary, i * epoch_step + j)
                if j % 50 == 0:
                    loss_value = \
                        sess.run([train_loss, copy_loss],
                                 feed_dict={x: x_batch[:, 0:n_his + 1, :, :], keep_prob: 1.0})
                    print(
                        f'Epoch {i:2d}, Step {j:3d}: [{loss_value[0]:.3f}, {loss_value[1]:.3f}]')
            print(
                f'Epoch {i:2d} Training Time {time.time() - start_time:.3f}s')
            if (i + 1) % args.save == 0:
                model_save(sess, global_steps, 'STGCN', save_path=sum_path)
        writer.close()
    print('Training model finished!')


def model_test(inputs, batch_size, n_his, n_pred, inf_mode, load_path='./output/models/'):
    '''
    Load and test saved model from the checkpoint.
    :param inputs: instance of class Dataset, data source for test.
    :param batch_size: int, the size of batch.
    :param n_his: int, the length of historical records for training.
    :param n_pred: int, the length of prediction.
    :param inf_mode: str, test mode - 'merge / multi-step test' or 'separate / single-step test'.
    :param load_path: str, the path of loaded model.
    '''
    model_path = tf.train.get_checkpoint_state(load_path).model_checkpoint_path

    test_graph = tf.Graph()

    with test_graph.as_default():
        saver = tf.train.import_meta_graph(pjoin(f'{model_path}.meta'))

    with tf.Session(graph=test_graph) as test_sess:
        saver.restore(test_sess, tf.train.latest_checkpoint(load_path))
        print(f'>> Loading saved model from {model_path} ...')

        pred = test_graph.get_collection('y_pred')

        if inf_mode == 'sep':
            # for inference mode 'sep', the type of step index is int.
            step_idx = n_pred - 1
            tmp_idx = [step_idx]
        elif inf_mode == 'merge':
            # for inference mode 'merge', the type of step index is np.ndarray.
            step_idx = tmp_idx = np.arange(3, n_pred + 1, 3) - 1
        else:
            raise ValueError(f'ERROR: test mode "{inf_mode}" is not defined.')

        x_test = inputs.get_data('test')

        y_test, len_test = multi_pred(
            test_sess, pred, x_test, batch_size, n_his, n_pred, step_idx)

    return y_test


def data_gen(data_loader):
    '''
    Source file load and dataset generation.
    :param file_path: str, the file path of data source.
    :param data_config: tuple, the configs of dataset in train, validation, test.
    :param n_route: int, the number of routes in the graph.
    :param n_frame: int, the number of frame within a standard sequence unit,
                         which contains n_his = 12 and n_pred = 9 (3 /15 min, 6 /30 min & 9 /45 min).
    :param day_slot: int, the number of time slots per day, controlled by the time window (5 min as default).
    :return: dict, dataset that contains training, validation and test with stats.
    '''

    # split data
    train_closeness, val_closeness = SplitData.split_data(
        data_loader.train_closeness, [0.9, 0.1])
    train_period, val_period = SplitData.split_data(
        data_loader.train_period, [0.9, 0.1])
    train_trend, val_trend = SplitData.split_data(
        data_loader.train_trend, [0.9, 0.1])
    train_y, val_y = SplitData.split_data(data_loader.train_y, [0.9, 0.1])
    test_y = data_loader.test_y[...,np.newaxis]
    train_y = train_y[...,np.newaxis]
    val_y = val_y[...,np.newaxis]
    # T, num_node, dimension, 1 -> T, dimension, num_node, 1
    seq_train = np.concatenate(
            [train_trend, train_period, train_closeness, train_y], axis=2).transpose([0, 2, 1, 3])
    seq_val = np.concatenate(
            [val_trend, val_period, val_closeness, val_y], axis=2).transpose([0, 2, 1, 3])
    seq_test = np.concatenate([data_loader.test_trend, data_loader.test_period, data_loader.test_closeness, test_y],
                                  axis=2).transpose([0, 2, 1, 3])

    # x_train, x_val, x_test: np.array, [sample_size, n_frame, n_route, channel_size].
    x_train = seq_train
    x_val =seq_val
    x_test = seq_test

    x_data = {'train': x_train, 'val': x_val, 'test': x_test}
    dataset = Dataset(x_data)
    return dataset


def multi_pred(sess, y_pred, seq, batch_size, n_his, n_pred, step_idx, dynamic_batch=True):
    '''
    Multi_prediction function.
    :param sess: tf.Session().
    :param y_pred: placeholder.
    :param seq: np.ndarray, [len_seq, n_frame, n_route, C_0].
    :param batch_size: int, the size of batch.
    :param n_his: int, size of historical records for training.
    :param n_pred: int, the length of prediction.
    :param step_idx: int or list, index for prediction slice.
    :param dynamic_batch: bool, whether changes the batch size in the last one if its length is less than the default.
    :return y_ : tensor, 'sep' [len_inputs, n_route, 1]; 'merge' [step_idx, len_inputs, n_route, 1].
            len_ : int, the length of prediction.
    '''
    pred_list = []
    for i in gen_batch(seq, min(batch_size, len(seq)), dynamic_batch=dynamic_batch):
        # Note: use np.copy() to avoid the modification of source data.
        test_seq = np.copy(i[:, 0:n_his + 1, :, :])
        step_list = []
        for j in range(n_pred):
            pred = sess.run(y_pred,
                            feed_dict={'data_input:0': test_seq, 'keep_prob:0': 1.0})
            if isinstance(pred, list):
                pred = np.array(pred[0])
            test_seq[:, 0:n_his - 1, :, :] = test_seq[:, 1:n_his, :, :]
            test_seq[:, n_his - 1, :, :] = pred
            step_list.append(pred)
        pred_list.append(step_list)
    #  pred_array -> [n_pred, batch_size, n_route, C_0)
    pred_array = np.concatenate(pred_list, axis=1)
    return pred_array[step_idx], pred_array.shape[1]


class Dataset(object):
    def __init__(self, data):
        self.__data = data


    def get_data(self, type):
        return self.__data[type]



    def get_len(self, type):
        return len(self.__data[type])



import numpy as np

import warnings
warnings.filterwarnings("ignore")


class HM(object):
    '''
    Historical Mean. A naive method that simply return average of hisrory data of each time slot.

    Args:
        c(int): The number of time slots of closeness history. 
        p (int): The number of time slots of period history which presents daily feature.
        t (int): The number of time slots of trend history which presents weekly feature.
        Note that `(c, p, t)` cannot all be zero at the same time. They denote how many
        features should be considerd in average.
    '''
    def __init__(self, c, p, t):

        self.c = c
        self.p = p
        self.t = t

        if self.c == 0 and self.p == 0 and self.t == 0:
            raise ValueError('c p t cannot all be zero at the same time')

    def predict(self, closeness_feature, period_feature, trend_feature):
        '''
        Give closeness, period and trend history values and then use their averages as predict.
        '''
        prediction = []

        if self.c > 0:
            prediction.append(closeness_feature[:, :, :, 0])

        if self.p > 0:
            prediction.append(period_feature[:, :, :, 0])

        if self.t > 0:
            prediction.append(trend_feature[:, :, :, 0])

        prediction = np.mean(np.concatenate(prediction, axis=-1), axis=-1, keepdims=True)

        return prediction



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="UCTB",
    version="0.3.5",
    author="UCTB group",
    author_email="chenliyue2019@gmail.com",
    description="Urban Computing ToolBox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/uctb/UCTB",
    packages=setuptools.find_packages(),
    install_requires=['hmmlearn>=0.2.1',
                      'keras==2.2.4',
                      'GPUtil',
                      'numpy>=1.16.2',
                      'pandas>=0.24.2',
                      'python-dateutil',
                      'scikit-learn>=0.20.3',
                      'scipy==1.2.1',
                      'statsmodels>=0.9.0',
                      'wget>=3.2',
                      'xgboost>=0.82',
                      'nni>=0.8',
                      'chinesecalendar>=1.2.2',
                      'workalendar>=8.2.0',
                      'PyYAML>=5.1.1'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import time
import math
import os
from UCTB.model.GMAN import GMAN
from gensim.models import Word2Vec
import datetime
from UCTB.preprocess import SplitData
from UCTB.train.LossFunction import mae_loss
import tensorflow as tf
import networkx as nx
import numpy as np


def build_model(log, time_fitness, trainX, args, SE):
    log_string(log, 'compiling model...')
    T = time_fitness
    print("time_fitness: ", T)
    num_train, _, N = trainX.shape
    X, TE, label, is_training = placeholder(args.P, args.Q, N)
    global_step = tf.Variable(0, trainable=False)
    bn_momentum = tf.compat.v1.train.exponential_decay(
        0.5, global_step,
        decay_steps=args.decay_epoch * num_train // args.batch_size,
        decay_rate=0.5, staircase=True)
    bn_decay = tf.minimum(0.99, 1 - bn_momentum)

    pred = GMAN(
        X, TE, SE, args.P, args.Q, T, args.L, args.K, args.d,
        bn=True, bn_decay=bn_decay, is_training=is_training)
    
    loss = mae_loss(pred, label)
    tf.compat.v1.add_to_collection('pred', pred)
    tf.compat.v1.add_to_collection('loss', loss)
    learning_rate = tf.compat.v1.train.exponential_decay(
        args.learning_rate, global_step,
        decay_steps=args.decay_epoch * num_train // args.batch_size,
        decay_rate=0.7, staircase=True)
    learning_rate = tf.maximum(learning_rate, 1e-5)
    optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate)
    train_op = optimizer.minimize(loss, global_step=global_step)
    parameters = 0
    for variable in tf.compat.v1.trainable_variables():
        parameters += np.product([x.value for x in variable.get_shape()])
    log_string(log, 'trainable parameters: {:,}'.format(parameters))
    log_string(log, 'model compiled!')
    saver = tf.compat.v1.train.Saver()
    config = tf.compat.v1.ConfigProto()
    config.gpu_options.allow_growth = True
    sess = tf.compat.v1.Session(config=config)
    sess.run(tf.compat.v1.global_variables_initializer())

    return X, TE, label, is_training, saver, sess, train_op, loss, pred


def Train(log, args, trainX, trainY, trainTE, valX, valTE, valY, X, TE, label, is_training, saver, sess, train_op, loss, pred):

    log_string(log, '**** training model ****')
    num_train, _, N = trainX.shape
    num_val = valX.shape[0]
    wait = 0
    val_loss_min = np.inf
    for epoch in range(args.max_epoch):
        if wait >= args.patience:
            log_string(log, 'early stop at epoch: %04d' % (epoch))
            break
        # shuffle
        permutation = np.random.permutation(num_train)
        trainX = trainX[permutation]
        trainTE = trainTE[permutation]
        trainY = trainY[permutation]
        # train loss
        start_train = time.time()
        train_loss = 0
        num_batch = math.ceil(num_train / args.batch_size)
        for batch_idx in range(num_batch):
            start_idx = batch_idx * args.batch_size
            end_idx = min(num_train, (batch_idx + 1) * args.batch_size)
            feed_dict = {
                X: trainX[start_idx: end_idx],
                TE: trainTE[start_idx: end_idx],
                label: trainY[start_idx: end_idx],
                is_training: True}
            # print(train_op.size())
            _, loss_batch = sess.run([train_op, loss], feed_dict=feed_dict)
            train_loss += loss_batch * (end_idx - start_idx)
        train_loss /= num_train
        end_train = time.time()
        # val loss
        start_val = time.time()
        val_loss = 0
        num_batch = math.ceil(num_val / args.batch_size)
        for batch_idx in range(num_batch):
            start_idx = batch_idx * args.batch_size
            end_idx = min(num_val, (batch_idx + 1) * args.batch_size)
            feed_dict = {
                X: valX[start_idx: end_idx],
                TE: valTE[start_idx: end_idx],
                label: valY[start_idx: end_idx],
                is_training: False}
            loss_batch = sess.run(loss, feed_dict=feed_dict)
            val_loss += loss_batch * (end_idx - start_idx)
        val_loss /= num_val
        end_val = time.time()
        log_string(
            log,
            '%s | epoch: %04d/%d, training time: %.1fs, inference time: %.1fs' %
            (datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), epoch + 1,
             args.max_epoch, end_train - start_train, end_val - start_val))
        log_string(
            log, 'train loss: %.4f, val_loss: %.4f' % (train_loss, val_loss))
        if val_loss <= val_loss_min:
            log_string(
                log,
                'val loss decrease from %.4f to %.4f, saving model to %s' %
                (val_loss_min, val_loss, args.model_file))
            wait = 0
            val_loss_min = val_loss
            saver.save(sess, args.model_file)
        else:
            wait += 1
    log_string(log, '**** testing model ****')
    log_string(log, 'loading model from %s' % args.model_file)
    saver = tf.compat.v1.train.import_meta_graph(args.model_file + '.meta')
    saver.restore(sess, args.model_file)
    log_string(log, 'model restored!')
    log_string(log, 'evaluating...')

    trainPred = []
    num_batch = math.ceil(num_train / args.batch_size)
    for batch_idx in range(num_batch):
        start_idx = batch_idx * args.batch_size
        end_idx = min(num_train, (batch_idx + 1) * args.batch_size)
        feed_dict = {
            X: trainX[start_idx: end_idx],
            TE: trainTE[start_idx: end_idx],
            is_training: False}
        pred_batch = sess.run(pred, feed_dict=feed_dict)
        trainPred.append(pred_batch)
    trainPred = np.concatenate(trainPred, axis=0)

    valPred = []
    num_batch = math.ceil(num_val / args.batch_size)
    for batch_idx in range(num_batch):
        start_idx = batch_idx * args.batch_size
        end_idx = min(num_val, (batch_idx + 1) * args.batch_size)
        feed_dict = {
            X: valX[start_idx: end_idx],
            TE: valTE[start_idx: end_idx],
            is_training: False}
        pred_batch = sess.run(pred, feed_dict=feed_dict)
        valPred.append(pred_batch)
    valPred = np.concatenate(valPred, axis=0)

    return trainPred, valPred


def Test(log, args, testX, testTE, X, TE, is_training, sess, pred):
    num_test = testX.shape[0]
    testPred = []
    num_batch = math.ceil(num_test / args.batch_size)
    start_test = time.time()
    for batch_idx in range(num_batch):
        start_idx = batch_idx * args.batch_size
        end_idx = min(num_test, (batch_idx + 1) * args.batch_size)
        feed_dict = {
            X: testX[start_idx: end_idx],
            TE: testTE[start_idx: end_idx],
            is_training: False}
        pred_batch = sess.run(pred, feed_dict=feed_dict)
        testPred.append(pred_batch)
    end_test = time.time()
    testPred = np.concatenate(testPred, axis=0)

    log_string(log, 'Test time: %.1fmin' % ((end_test - start_test) / 60))
    sess.close()
    log.close()

    return testPred


def load_data(args, data_loader):
    # split data
    train_closeness, val_closeness = SplitData.split_data(
        data_loader.train_closeness, [0.9, 0.1])
    train_period, val_period = SplitData.split_data(
        data_loader.train_period, [0.9, 0.1])
    train_trend, val_trend = SplitData.split_data(
        data_loader.train_trend, [0.9, 0.1])
    train_y, val_y = SplitData.split_data(data_loader.train_y, [0.9, 0.1])
    print(train_y.shape)

    trainY = train_y.reshape([train_y.shape[0], 1, data_loader.station_number])
    valY = val_y.reshape([val_y.shape[0], 1, data_loader.station_number])
    testY = data_loader.test_y.reshape(
        [data_loader.test_y.shape[0], 1, data_loader.station_number])

    # T, num_node, dimension -> T, dimension, num_node
    if data_loader.period_len > 0 and data_loader.trend_len > 0:
        trainX = np.concatenate(
            [train_trend, train_period, train_closeness], axis=2).squeeze().transpose([0, 2, 1])
        print(trainX.shape)
        valX = np.concatenate(
            [val_trend, val_period, val_closeness], axis=2).squeeze().transpose([0, 2, 1])
        testX = np.concatenate([data_loader.test_trend, data_loader.test_period, data_loader.test_closeness],
                               axis=2).squeeze().transpose([0, 2, 1])
    else:
        trainX = train_closeness.squeeze().transpose([0, 2, 1])
        valX = val_closeness.squeeze().transpose([0, 2, 1])
        testX = data_loader.test_closeness.squeeze().transpose([0, 2, 1])


    # spatial embedding
    f = open(args.SE_file, mode='r')
    lines = f.readlines()
    temp = lines[0].split(' ')
    N, dims = int(temp[0]), int(temp[1])
    SE = np.zeros(shape=(N, dims), dtype=np.float32)
    for line in lines[1:]:
        temp = line.split(' ')
        index = int(temp[0])
        SE[index] = temp[1:]

    train_steps, val_steps, test_steps = len(trainX), len(valX), len(testX)
    # print("train_steps",train_steps)
    # print("test_steps",test_steps)
    # print("val_steps",val_steps)

    # temporal embedding
    time_fitness = data_loader.dataset.time_fitness
    time_delta = datetime.timedelta(minutes=time_fitness)
    try:
        start_time = datetime.datetime.strptime(
            data_loader.dataset.time_range[0], "%Y-%m-%d")
    except:
        start_time = datetime.datetime.strptime(
            data_loader.dataset.time_range[0], "%Y-%m-%d %H:%M:%S")

    Time = [start_time + i *
            time_delta for i in range(train_steps + test_steps + val_steps + args.P)]

    def get_attr(object, attr):
        return np.array([eval("item.{}".format(attr)) for item in object])

    print(data_loader.dataset.time_range[0])
    dayofweek = np.reshape(get_attr(Time, "weekday()"), newshape=(-1, 1))
    # timeofday = (Time.hour * 3600 + Time.minute * 60 + Time.second) \
    #             // Time.freq.delta.total_seconds()
    timeofday = (get_attr(Time, "hour") * 3600 + get_attr(Time, "minute") * 60 + get_attr(Time, "second")) \
        // (86400)
    timeofday = np.reshape(timeofday, newshape=(-1, 1))
    Time = np.concatenate((dayofweek, timeofday), axis=-1)
    # train/val/test
    train = Time[: train_steps + args.P]
    val = Time[train_steps: train_steps + val_steps + args.P]
    test = Time[train_steps + val_steps:]
    # shape = (num_sample, P + Q, 2)
    trainTE = seq2instance(train, args.P, args.Q)
    trainTE = np.concatenate(trainTE, axis=1).astype(np.int32)
    valTE = seq2instance(val, args.P, args.Q)
    valTE = np.concatenate(valTE, axis=1).astype(np.int32)
    testTE = seq2instance(test, args.P, args.Q)
    testTE = np.concatenate(testTE, axis=1).astype(np.int32)
    
    return (trainX, trainTE, trainY, valX, valTE, valY, testX, testTE, testY,
            SE, time_fitness)


def placeholder(P, Q, N):
    X = tf.compat.v1.placeholder(shape=(None, P, N), dtype=tf.float32)
    TE = tf.compat.v1.placeholder(shape=(None, P + Q, 2), dtype=tf.int32)
    label = tf.compat.v1.placeholder(shape=(None, Q, N), dtype=tf.float32)
    is_training = tf.compat.v1.placeholder(shape=(), dtype=tf.bool)
    return X, TE, label, is_training


def graph_to_adj_files(adjacent_matrix, Adj_file):
    with open(Adj_file, "w") as fp:
        adj_list = []
        print(adjacent_matrix.shape)
        for i in range(adjacent_matrix.shape[0]):
            for j in range(adjacent_matrix.shape[1]):
                adj_list.append("{} {} {:.6f}\n".format(
                    i, j, adjacent_matrix[i, j]))
        fp.writelines(adj_list)


def read_graph(edgelist):
    G = nx.read_edgelist(
        edgelist, nodetype=int, data=(('weight', float),),
        create_using=nx.DiGraph())
    return G


def learn_embeddings(walks, dimensions, output_file, epochs):
    walks = [list(map(str, walk)) for walk in walks]
    model = Word2Vec(
        walks, vector_size=dimensions, window=10, min_count=0, sg=1,
        workers=8, epochs=epochs)
    model.wv.save_word2vec_format(output_file)
    return


def seq2instance(data, P, Q):
    num_step, dims = data.shape
    num_sample = num_step - P - Q + 1
    x = np.zeros(shape=(num_sample, P, dims))
    y = np.zeros(shape=(num_sample, Q, dims))
    for i in range(num_sample):
        x[i] = data[i: i + P]
        y[i] = data[i + P: i + P + Q]
    return x, y

# log string
def log_string(log, string):
    log.write(string + '\n')
    log.flush()
    print(string)



import torch
import torch.nn as nn
import torch.nn.functional as F

class nconv(nn.Module):
    def __init__(self):
        super(nconv,self).__init__()

    def forward(self,x, A):
        x = torch.einsum('ncvl,vw->ncwl',(x,A))
        return x.contiguous()

class linear(nn.Module):
    def __init__(self,c_in,c_out):
        super(linear,self).__init__()
        self.mlp = torch.nn.Conv2d(c_in, c_out, kernel_size=(1, 1), padding=(0,0), stride=(1,1), bias=True)

    def forward(self,x):
        return self.mlp(x)

class gcn(nn.Module):
    def __init__(self,c_in,c_out,dropout,support_len=3,order=2):
        super(gcn,self).__init__()
        self.nconv = nconv()
        c_in = (order*support_len+1)*c_in
        self.mlp = linear(c_in,c_out)
        self.dropout = dropout
        self.order = order

    def forward(self,x,support):
        out = [x]
        for a in support:
            x1 = self.nconv(x,a)
            out.append(x1)
            for k in range(2, self.order + 1):
                x2 = self.nconv(x1,a)
                out.append(x2)
                x1 = x2

        h = torch.cat(out,dim=1)
        h = self.mlp(h)
        h = F.dropout(h, self.dropout, training=self.training)
        return h


class gwnet(nn.Module):
    """

    References:
        - `Graph wavenet for deep spatial-temporal graph modeling.
          <https://www.ijcai.org/proceedings/2019/0264.pdf>`_.
        - `A PyTorch implementation of the GraphWaveNet model  (nnzhan)
          <https://github.com/nnzhan/Graph-WaveNet>`_.

    Args:
        device(torch.device): Which device use to train.
        num_node(int): Number of blocks.
        drop_out(int): Number of input channels.
        supports(int): Order of chebyshev polynomial.
        gcn_bool(int): Number of chebyshev filter.
        addaptadj(bool): Whether to add adaptive adjacent matrix.
        aptinit(torch.tensor): Initialization of adjacent matrix.
        in_dim(int): Number of input's dimension.
        out_dim(int): Number of output's dimension.
        residual_channels(int): Number of channels after residual module.
        dilation_channels(int): Number of channels after dilation module.
        skip_channels(int): Number of skip channels.
        end_channels(int): Number of end channels.
        kernel_size(int): Kernel Size for dilation convolution.
        blocks(int): Number of block.
        layers(int): Number of layer.
    """
    def __init__(self, device, num_node, dropout=0.3, supports=None, gcn_bool=True, addaptadj=True, aptinit=None, in_dim=2,out_dim=12,residual_channels=32,dilation_channels=32,skip_channels=256,end_channels=512,kernel_size=2,blocks=4,layers=2):
        super(gwnet, self).__init__()
        self.dropout = dropout
        self.blocks = blocks
        self.layers = layers
        self.gcn_bool = gcn_bool
        self.addaptadj = addaptadj

        self.filter_convs = nn.ModuleList()
        self.gate_convs = nn.ModuleList()
        self.residual_convs = nn.ModuleList()
        self.skip_convs = nn.ModuleList()
        self.bn = nn.ModuleList()
        self.gconv = nn.ModuleList()

        self.start_conv = nn.Conv2d(in_channels=in_dim,
                                    out_channels=residual_channels,
                                    kernel_size=(1,1))
        self.supports = supports

        receptive_field = 1

        self.supports_len = 0
        if supports is not None:
            self.supports_len += len(supports)

        if gcn_bool and addaptadj:
            if aptinit is None:
                if supports is None:
                    self.supports = []
                self.nodevec1 = nn.Parameter(torch.randn(num_node, 10).to(device), requires_grad=True).to(device)
                self.nodevec2 = nn.Parameter(torch.randn(10, num_node).to(device), requires_grad=True).to(device)
                self.supports_len +=1
            else:
                if supports is None:
                    self.supports = []
                m, p, n = torch.svd(aptinit)
                initemb1 = torch.mm(m[:, :10], torch.diag(p[:10] ** 0.5))
                initemb2 = torch.mm(torch.diag(p[:10] ** 0.5), n[:, :10].t())
                self.nodevec1 = nn.Parameter(initemb1, requires_grad=True).to(device)
                self.nodevec2 = nn.Parameter(initemb2, requires_grad=True).to(device)
                self.supports_len += 1




        for b in range(blocks):
            additional_scope = kernel_size - 1
            new_dilation = 1
            for i in range(layers):
                # dilated convolutions
                self.filter_convs.append(nn.Conv2d(in_channels=residual_channels,
                                                   out_channels=dilation_channels,
                                                   kernel_size=(1,kernel_size),dilation=new_dilation))

                self.gate_convs.append(nn.Conv1d(in_channels=residual_channels,
                                                 out_channels=dilation_channels,
                                                 kernel_size=(1, kernel_size), dilation=new_dilation))

                # 1x1 convolution for residual connection
                self.residual_convs.append(nn.Conv1d(in_channels=dilation_channels,
                                                     out_channels=residual_channels,
                                                     kernel_size=(1, 1)))

                # 1x1 convolution for skip connection
                self.skip_convs.append(nn.Conv1d(in_channels=dilation_channels,
                                                 out_channels=skip_channels,
                                                 kernel_size=(1, 1)))
                self.bn.append(nn.BatchNorm2d(residual_channels))
                new_dilation *=2
                receptive_field += additional_scope
                additional_scope *= 2
                if self.gcn_bool:
                    self.gconv.append(gcn(dilation_channels,residual_channels,dropout,support_len=self.supports_len))



        self.end_conv_1 = nn.Conv2d(in_channels=skip_channels,
                                  out_channels=end_channels,
                                  kernel_size=(1,1),
                                  bias=True)

        self.end_conv_2 = nn.Conv2d(in_channels=end_channels,
                                    out_channels=out_dim,
                                    kernel_size=(1,1),
                                    bias=True)

        self.receptive_field = receptive_field



    def forward(self, input):
        in_len = input.size(3)
        if in_len<self.receptive_field:
            x = nn.functional.pad(input,(self.receptive_field-in_len,0,0,0))
        else:
            x = input
        x = self.start_conv(x)
        skip = 0

        # calculate the current adaptive adj matrix once per iteration
        new_supports = None
        if self.gcn_bool and self.addaptadj and self.supports is not None:
            adp = F.softmax(F.relu(torch.mm(self.nodevec1, self.nodevec2)), dim=1)
            new_supports = self.supports + [adp]

        # WaveNet layers
        for i in range(self.blocks * self.layers):

            #            |----------------------------------------|     *residual*
            #            |                                        |
            #            |    |-- conv -- tanh --|                |
            # -> dilate -|----|                  * ----|-- 1x1 -- + -->	*input*
            #                 |-- conv -- sigm --|     |
            #                                         1x1
            #                                          |
            # ---------------------------------------> + ------------->	*skip*

            #(dilation, init_dilation) = self.dilations[i]

            #residual = dilation_func(x, dilation, init_dilation, i)
            residual = x
            # dilated convolution
            filter = self.filter_convs[i](residual)
            filter = torch.tanh(filter)
            gate = self.gate_convs[i](residual)
            gate = torch.sigmoid(gate)
            x = filter * gate

            # parametrized skip connection

            s = x
            s = self.skip_convs[i](s)
            try:
                skip = skip[:, :, :,  -s.size(3):]
            except:
                skip = 0
            skip = s + skip


            if self.gcn_bool and self.supports is not None:
                if self.addaptadj:
                    x = self.gconv[i](x, new_supports)
                else:
                    x = self.gconv[i](x,self.supports)
            else:
                x = self.residual_convs[i](x)

            x = x + residual[:, :, :, -x.size(3):]


            x = self.bn[i](x)

        x = F.relu(skip)
        x = F.relu(self.end_conv_1(x))
        x = self.end_conv_2(x)
        return x


import numpy as np
from hmmlearn import hmm


class HMM(object):
    def __init__(self, num_components, n_iter, hmm_kernal=hmm.GaussianHMM):

        self._num_components = num_components
        self._iter = n_iter

        self._hmm = hmm_kernal(n_components=self._num_components, n_iter=self._iter, covariance_type='full')

    def fit(self, x):
        self._hmm.fit(x)
        if self._hmm.monitor_.converged:
            print('Status: converged')

    def predict(self, x, length):
        # predict the state for each element of X
        # and store the last state
        last_state = self._hmm.predict_proba(x)[-1:]

        pre_state = []
        pre_observation = []

        for i in range(length):
            # predict the state of next moment using the transmat
            last_state = np.dot(last_state, self._hmm.transmat_)

            pre_state.append(last_state)

            # dot product between the state-probability and state-means
            pre_observation.append([np.dot(last_state, self._hmm.means_)[0][0]])

        return pre_observation